"""
Connector Base Method
=============================

This module defines the base class for connector functions that allow the combination of 
agents' observations to provide a flexible configuration of the communication between agents. 
Built-in hierarchical (only two levels), fully-shared, centralized, and independent configurations 
are provided.
"""
from typing import Dict, Any, Tuple, List
from gymnasium import spaces

from eprllib import logger
from eprllib.Utils.annotations import OverrideToImplementCustomLogic

class BaseConnector:
    """
    Base class for connector functions.
    """
    connector_fn_config: Dict[str, Any]
    obs_indexed: Dict[str,Dict[str, int]]
    _is_setup: bool
    
    def __init__(
        self,
        connector_fn_config: Dict[str, Any] = {}
    ):
        """
        Base class for connector functions.
        
        Args:
            connector_fn_config (Dict[str, Any]): Configuration of the connector function.
        
        Raises:
            RuntimeError: If the ``setup()`` method is called more than once.
            AttributeError: If the ``setup()`` method is not implemented in the child class.
        """
        self.connector_fn_config = connector_fn_config
        self.obs_indexed = {}
        
        logger.info(f"BaseConnector: The BaseConnector was correctly inicializated with {self.connector_fn_config} config.")
        
        # Make sure, `setup()` is only called once, no matter what.
        if hasattr(self, "_is_setup") and self._is_setup:
            raise RuntimeError(
                "``BaseActionMapper.setup()`` called twice within your ActionMapper implementation "
                f"{self}! Make sure you are using the proper inheritance order "
                " and that you are NOT overriding the constructor, but "
                "only the ``setup()`` method of your subclass."
            )
        try:
            self.setup()
        except AttributeError as e:
            raise e

        self._is_setup:bool = True
        
    
    def __name__(self):
        """
        Returns the name of the connector function.
        
        Returns:
            str: Name of the connector function.
        """
        return self.__class__.__name__
    
    def get_all_agents_obs_spaces_dict(
        self,
        possible_agents: List[str]
    ) -> spaces.Dict:
        """
        Get all the agents observations spaces putting togheter in a Dict space dimension.
        
        Args:
            possible_agents (List[str]): List of possible agents.
        
        Returns:
            gym.spaces.Dict: Agents observation spaces.
        """
        observation_space_dict: Dict[str, Any] = {}
        for agent in possible_agents:
            observation_space_dict[agent] = self.get_agent_obs_dim(agent)
        return spaces.Dict(observation_space_dict)
    
    
    # ===========================
    # === OVERRIDABLE METHODS ===
    # ===========================
    
    @OverrideToImplementCustomLogic
    def setup(self):
        """
        This method can be overridden in subclasses to perform setup tasks.
        """
        pass
    
    @OverrideToImplementCustomLogic
    def get_agent_obs_dim(
        self,
        agent: str
    ) -> spaces.Space[Any]:
        """
        Get the agent observation dimension.
        
        Args:
            agent (str): Agent identifier.
        
        Returns:
            gym.spaces.Space: Agent observation dimension.
            
        Raises:
            NotImplementedError: If the method is not implemented in the child class.
        """
        msg = "BaseConnector: This method must be implemented in the child class."
        logger.error(msg)
        raise NotImplementedError(msg)
    
    @OverrideToImplementCustomLogic
    def get_agent_obs_indexed(
        self,
        agent: str
    ) -> Dict[str, int]:
        """
        Get a dictionary of the agent observation parameters and their respective index in the observation array.
        
        Args:
            agent (str): Agent identifier.
            
        Returns:
            Dict[str, int]: Dictionary of the agent observation parameters and their respective index in the observation array.
        
        Raises:
            NotImplementedError: If the method is not implemented in the child class.

        """
        msg = "BaseConnector: This method must be implemented in the child class."
        logger.error(msg)
        raise NotImplementedError(msg)
    
    
    def set_top_level_obs(
        self,
        agent_states: Dict[str, Dict[str, Any]],
        dict_agents_obs: Dict[str, Any],
        infos: Dict[str, Dict[str, Any]],
        is_last_timestep: bool = False
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """
        Set the multi-agent observation.
        
        Args:
            agent_states (Dict[str, Dict[str, Any]]): Agent states.
            dict_agents_obs (Dict[str, Any]): Dictionary of agents' observations.
            infos (Dict[str, Dict[str, Any]]): Additional information.
            is_last_timestep (bool, optional): Flag indicating if it is the last timestep. Defaults to False.
        
        Returns:
            Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]: Multi-agent observation, updated infos, and a flag indicating if it is the lowest level.
        
        """
        is_lowest_level = True
        return dict_agents_obs, infos, is_lowest_level
    
    
    def set_low_level_obs(
        self,
        agent_states: Dict[str,Dict[str,Any]],
        dict_agents_obs: Dict[str,Any],
        infos: Dict[str, Dict[str, Any]],
        goals: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]:
        """
        Set the multiagent observation.
        
        Args:
            agent_states (Dict[str, Dict[str, Any]]): Agent states.
            dict_agents_obs (Dict[str, Any]): Dictionary of agents' observations.
            infos (Dict[str, Dict[str, Any]]): Additional information.
            goals (Dict[str, Any]): Goals.
        
        Returns:
            Tuple[Dict[str, Any], Dict[str, Dict[str, Any]], bool]: Multi-agent observation, updated infos, and a flag indicating if it is the lowest level.
            
        """
        is_lowest_level = True
        return dict_agents_obs, infos, is_lowest_level



"""
Environment Utilities
=====================

"""
from typing import List, Dict
from eprllib.Agents import OCCUPATION_PROFILES
from numpy import random, exp
from datetime import datetime, timedelta

def _calculate_occupancy_once(
    current_hour: int, 
    current_day_type: int,
    current_holiday: bool, 
    user_type: str, 
    zone_type: str
) -> float:
    """
    Internal function that calculates the number of occupants for a single stochastic simulation.
    """
    profile = OCCUPATION_PROFILES[user_type]
    # Selecting the correct zone profile
    profile_zone: Dict[str, List[float]] = profile[f"zone_{zone_type}"]
    # Selecting the correct type of day
    if current_holiday or current_day_type in [5, 6]:
        base_schedule = profile_zone["weekends"]
    else:
        base_schedule = profile_zone["weekdays"]
    # Selecting the occupation based on the current time
    base_occupation = base_schedule[current_hour]
    
    return base_occupation


def calculate_occupancy(
    current_hour: int,
    current_day: int,
    current_month: int,
    current_year: int,
    current_holiday: bool,
    user_type: str,
    zone_type: str,
    confidence_level: float = 0.95
) -> float:
    """
    Calculates current occupancy and a probability forecast for a specific area type.

    Args:
        current_hour (int): Current time (0-23).
        current_day (int): Day of the month (1-31).
        current_month (int): Month of the year (1-12).
        current_year (int): Current year (e.g., 2023).
        current_holiday (bool): True if the current day is a holiday.
        user_type (str): The occupancy profile to use.
        zone_type (str): The type of zone to simulate ('day' or 'night').
        confidence_level (float): Confidence that the occupancy is the same that the deterministic.

    Returns:
        tuple[int, list[float]]: Una tupla conteniendo:
            - int: El número estimado de personas presentes en la hora actual.
            - list[float]: Una lista con 24 valores de probabilidad (0-1) de ocupación.
    """
    current_time_obj = datetime(current_year, current_month, current_day, current_hour)
    current_day_type = current_time_obj.weekday()

    current_occupation = _calculate_occupancy_once(
        current_hour = current_hour, 
        current_day_type = current_day_type,
        current_holiday = current_holiday, 
        user_type = user_type, 
        zone_type = zone_type
    )
    if current_occupation > 0:
        if random.random() > confidence_level:
            current_occupation = 0.
    else:
        if random.random() > confidence_level:
            current_occupation = 1.
    
    return current_occupation

def calculate_occupancy_forecast(
    current_hour: int,
    current_day: int,
    current_month: int,
    current_year: int,
    user_type: str,
    zone_type: str,
    occupation_prediction_hours: int = 24,
    confidence_level: float = 0.95,
    lambdaa: float = 0.05
) -> List[float]:
    """
    Calculates current occupancy and a probability forecast for a specific area type.

    Args:
        current_hour (int): Current time (0-23).
        current_day (int): Day of the month (1-31).
        current_month (int): Month of the year (1-12).
        current_year (int): Current year (e.g., 2023).
        current_holiday (bool): True if the current day is a holiday.
        user_type (str): The occupancy profile to use.
        zone_type (str): The type of zone to simulate ('day' or 'night').
        confidence_level (float): Confidence that the occupancy is the same that the deterministic.
        lambdaa (float): Decay value of the confidence level.

    Returns:
        tuple[int, list[float]]: Una tupla conteniendo:
            - int: El número estimado de personas presentes en la hora actual.
            - list[float]: Una lista con 24 valores de probabilidad (0-1) de ocupación.
    """
    current_time_obj = datetime(current_year, current_month, current_day, current_hour)
    
    forecast_probabilities: List[float] = []
    
    for h in range(1, occupation_prediction_hours+1):
        confidence_level_h = 0.5+(confidence_level-0.5)*exp(-lambdaa*h)
        future_time = current_time_obj + timedelta(hours=h)
        is_a_future_holiday = False  # Simplification for forecasting
        
        simulated_occupants = _calculate_occupancy_once(
            current_hour = future_time.hour, 
            current_day_type = future_time.weekday(),
            current_holiday = is_a_future_holiday, 
            user_type = user_type, 
            zone_type = zone_type
        )
        if simulated_occupants > 0:
            simulated_occupants = confidence_level_h
        else:
            simulated_occupants = 1 - confidence_level_h
            
        forecast_probabilities.append(simulated_occupants)
    
    return forecast_probabilities



"""
Agent utilities
================

"""
from typing import Any, Dict, List, get_args, get_origin
from eprllib import logger

def get_agent_name(state: Dict[str, Any]|List[str]) -> str:
    """
    Get the agent name from the state dictionary. The state dictionay can be the agent_states dictionary or the infos 
    dictionary. Both contain the agent name in all the variables keys as: "agent_name: variable[0]: variable[1]: ...".
    This method is util for autodetect the agent name in the implementation of filters, ActionMappers, or rewards classes for
    agents.

    Args:
        state (Dict[str, Any] | List): The state dictionary.

    Returns:
        str: The agent name.
    
    Raises:
        TypeError: If the state is neither a dictionary nor a list.
        ValueError: If the state is empty.
    """
    # As we don't know the agent that belong this filter, we auto-dectect his name form the name of the variables names
    # inside the agent_states_copy dictionary. The agent_states dict has keys with the format of "agent_name: ...".
    if isinstance(state, List):
        # Check if the list is empty
        if not state:
            msg = "AgentUtils: The state list is empty"
            logger.error(msg)
            raise ValueError(msg)
        return state[0].split(':')[0]
    elif isinstance(state, Dict):
        # Check if the dictionary is empty
        if not state:
            msg = "AgentUtils: The state dictionary is empty"
            logger.error(msg)
            raise ValueError(msg)
        return list(state.keys())[0].split(':')[0]
    else:
        msg = f"AgentUtils: The state must be a dictionary or a list, but got a {type(state).__name__}"
        logger.error(msg)
        raise TypeError(msg)




def _validate_type(value: Any, expected_type: Any, path: str = "") -> None:
    """Recursively validates a value against an expected type."""
    origin = get_origin(expected_type)
    
    if origin is tuple:
        # if the origin is a list, try to convert in a tuple
        if isinstance(value, list):
            value = tuple(value)
        
        elif not isinstance(value, tuple):
            msg = f"AgentUtils: Expected tuple at {path}, got {type(value).__name__}"
            logger.error(msg)
            raise TypeError(msg)
        
        expected_types = get_args(expected_type)
        if len(value) != len(expected_types):
            msg = f"AgentUtils: Expected tuple with {len(expected_types)} elements at {path}, got {len(value)}"
            logger.error(msg)
            raise ValueError(msg)
        
        for i, (item, item_type) in enumerate(zip(value, expected_types)):
            _validate_type(item, item_type, f"{path}[{i}]")
    
    elif origin is dict:
        if not isinstance(value, dict):
            msg = f"AgentUtils: Expected dict at {path}, got {type(value).__name__}"
            logger.error(msg)
            raise TypeError(msg)
        
        key_type, value_type = get_args(expected_type)
        for k, v in value.items():
            _validate_type(k, key_type, f"{path}.key")
            _validate_type(v, value_type, f"{path}[{k}]")
    
    elif origin is list:
        if not isinstance(value, list):
            msg = f"AgentUtils: Expected list at {path}, got {type(value).__name__}"
            logger.error(msg)
            raise TypeError(msg)
        
        item_type = get_args(expected_type)[0]
        for i, item in enumerate(value):
            _validate_type(item, item_type, f"{path}[{i}]")
    
    elif expected_type is Any:
        pass  # Any type is always valid
    
    elif not isinstance(value, expected_type):
        msg = f"AgentUtils: Expected {expected_type.__name__} at {path}, got {type(value).__name__}"
        logger.error(msg)
        raise TypeError(msg)


def config_validation(config: Dict[str, Any], required_keys: Dict[str, Any]) -> None:
    """
    Validates a configuration dictionary against required keys and nested types.

    Args:
        config: The configuration dictionary to validate.
        required_keys: A dictionary where keys are required key names and values are
                       expected types (supports nested types like Dict[str, Tuple[int, float, Any]]).

    Raises:
        ValueError: If a required key is missing.
        TypeError: If a value has an incorrect type.
    """
    for key, expected_type in required_keys.items():
        if key not in config:
            msg = f"AgentUtils: Missing required key: '{key}'"
            logger.error(msg)
            raise ValueError(msg)
        
        _validate_type(config[key], expected_type, f"config['{key}']")



"""
This type stub file was generated by pyright.
"""

import collections
import attr
import wrapt
from typing import Any, Mapping
from absl.testing import parameterized

"""Tests for utilities working with arbitrarily nested structures."""
STRUCTURE1 = ...
STRUCTURE2 = ...
STRUCTURE_DIFFERENT_NUM_ELEMENTS = ...
STRUCTURE_DIFFERENT_NESTING = ...
class DoctestTest(parameterized.TestCase):
  def testDoctest(self): # -> None:
    ...
  


class NestTest(parameterized.TestCase):
  def assertAllEquals(self, a, b): # -> None:
    ...
  
  def testAttrsFlattenAndUnflatten(self): # -> None:
    class BadAttr:
      """Class that has a non-iterable __attrs_attrs__."""
      ...
    
    
    @attr.s
    class SampleAttr:
      ...
    
    
  
  @parameterized.parameters([(1, 2, 3), ({ "B": 10,"A": 20 }, [1, 2], 3), ((1, 2), [3, 4], 5), (collections.namedtuple("Point", ["x", "y"])(1, 2), 3, 4), wrapt.ObjectProxy((collections.namedtuple("Point", ["x", "y"])(1, 2), 3, 4))])
  def testAttrsMapStructure(self, *field_values): # -> None:
    @attr.s
    class SampleAttr:
      ...
    
    
  
  def testFlattenAndUnflatten(self): # -> None:
    ...
  
  def testFlattenDictOrder(self): # -> None:
    ...
  
  def testUnflattenDictOrder(self): # -> None:
    ...
  
  def testFlattenAndUnflatten_withDicts(self): # -> None:
    ...
  
  def testFlatten_numpyIsNotFlattened(self): # -> None:
    ...
  
  def testFlatten_stringIsNotFlattened(self): # -> None:
    ...
  
  def testFlatten_bytearrayIsNotFlattened(self): # -> None:
    ...
  
  def testUnflattenSequenceAs_notIterableError(self): # -> None:
    ...
  
  def testUnflattenSequenceAs_wrongLengthsError(self): # -> None:
    ...
  
  def testUnflattenSequenceAs_defaultdict(self): # -> None:
    ...
  
  def testIsSequence(self): # -> None:
    ...
  
  Named0ab = ...
  Named1ab = ...
  SameNameab = ...
  SameNameab2 = ...
  SameNamexy = ...
  SameName1xy = ...
  SameName1xy2 = ...
  NotSameName = ...
  class SameNamedType1(SameNameab):
    ...
  
  
  def testAssertSameStructure(self): # -> None:
    ...
  
  def testAssertSameStructure_differentNumElements(self): # -> None:
    ...
  
  def testAssertSameStructure_listVsNdArray(self): # -> None:
    ...
  
  def testAssertSameStructure_intVsList(self): # -> None:
    ...
  
  def testAssertSameStructure_tupleVsList(self): # -> None:
    ...
  
  def testAssertSameStructure_differentNesting(self): # -> None:
    ...
  
  def testAssertSameStructure_tupleVsNamedTuple(self): # -> None:
    ...
  
  def testAssertSameStructure_sameNamedTupleDifferentContents(self): # -> None:
    ...
  
  def testAssertSameStructure_differentNamedTuples(self): # -> None:
    ...
  
  def testAssertSameStructure_sameNamedTupleDifferentStructuredContents(self): # -> None:
    ...
  
  def testAssertSameStructure_differentlyNestedLists(self): # -> None:
    ...
  
  def testAssertSameStructure_listStructureWithAndWithoutTypes(self): # -> None:
    ...
  
  def testAssertSameStructure_dictionaryDifferentKeys(self): # -> None:
    ...
  
  def testAssertSameStructure_sameNameNamedTuples(self): # -> None:
    ...
  
  def testAssertSameStructure_sameNameNamedTuplesNested(self): # -> None:
    ...
  
  def testAssertSameStructure_sameNameNamedTuplesDifferentStructure(self): # -> None:
    ...
  
  def testAssertSameStructure_differentNameNamedStructures(self): # -> None:
    ...
  
  def testAssertSameStructure_sameNameDifferentFieldNames(self): # -> None:
    ...
  
  def testAssertSameStructure_classWrappingNamedTuple(self): # -> None:
    ...
  
  def testMapStructure(self): # -> None:
    ...
  
  def testMapStructureWithStrings(self): # -> None:
    ...
  
  def testAssertShallowStructure(self): # -> None:
    ...
  
  def testFlattenUpTo(self): # -> None:
    @attr.s
    class ABAttr:
      ...
    
    
  
  def testByteStringsNotTreatedAsIterable(self): # -> None:
    ...
  
  def testFlattenWithPathUpTo(self): # -> None:
    ...
  
  def testMapStructureUpTo(self): # -> None:
    ...
  
  Foo = ...
  Bar = ...
  @parameterized.parameters([dict(inputs=[], expected=[]), dict(inputs=[23, "42"], expected=[((0, ), 23), ((1, ), "42")]), dict(inputs=[[[[108]]]], expected=[((0, 0, 0, 0), 108)]), dict(inputs=Foo(a=3, b=Bar(c=23, d=42)), expected=[(("a", ), 3), (("b", "c"), 23), (("b", "d"), 42)]), dict(inputs=Foo(a=Bar(c=23, d=42), b=Bar(c=0, d="thing")), expected=[(("a", "c"), 23), (("a", "d"), 42), (("b", "c"), 0), (("b", "d"), "thing")]), dict(inputs=Bar(c=42, d=43), expected=[(("c", ), 42), (("d", ), 43)]), dict(inputs=Bar(c=[42], d=43), expected=[(("c", 0), 42), (("d", ), 43)]), dict(inputs=wrapt.ObjectProxy(Bar(c=[42], d=43)), expected=[(("c", 0), 42), (("d", ), 43)])])
  def testFlattenWithPath(self, inputs, expected): # -> None:
    ...
  
  @parameterized.named_parameters([dict(testcase_name="Tuples", s1=(1, 2), s2=(3, 4), check_types=True, expected=(((0, ), 4), ((1, ), 6))), dict(testcase_name="Dicts", s1={ "a": 1,"b": 2 }, s2={ "b": 4,"a": 3 }, check_types=True, expected={ "a": (("a", ), 4),"b": (("b", ), 6) }), dict(testcase_name="Mixed", s1=(1, 2), s2=[3, 4], check_types=False, expected=(((0, ), 4), ((1, ), 6))), dict(testcase_name="Nested", s1={ "a": [2, 3],"b": [1, 2, 3] }, s2={ "b": [5, 6, 7],"a": [8, 9] }, check_types=True, expected={ "a": [(("a", 0), 10), (("a", 1), 12)],"b": [(("b", 0), 6), (("b", 1), 8), (("b", 2), 10)] })])
  def testMapWithPathCompatibleStructures(self, s1, s2, check_types, expected): # -> None:
    ...
  
  @parameterized.named_parameters([dict(testcase_name="Tuples", s1=(1, 2, 3), s2=(4, 5), error_type=ValueError), dict(testcase_name="Dicts", s1={ "a": 1 }, s2={ "b": 2 }, error_type=ValueError), dict(testcase_name="Nested", s1={ "a": [2, 3, 4],"b": [1, 3] }, s2={ "b": [5, 6],"a": [8, 9] }, error_type=ValueError)])
  def testMapWithPathIncompatibleStructures(self, s1, s2, error_type): # -> None:
    ...
  
  def testMappingProxyType(self): # -> None:
    ...
  
  def testTraverseListsToTuples(self): # -> None:
    ...
  
  def testTraverseEarlyTermination(self): # -> None:
    ...
  
  def testMapStructureAcrossSubtreesDict(self): # -> None:
    ...
  
  def testMapStructureAcrossSubtreesNoneValues(self): # -> None:
    ...
  
  def testMapStructureAcrossSubtreesList(self): # -> None:
    ...
  
  def testMapStructureAcrossSubtreesTuple(self): # -> None:
    ...
  
  def testMapStructureAcrossSubtreesNamedTuple(self): # -> None:
    ...
  
  def testMapStructureAcrossSubtreesListTuple(self): # -> None:
    ...
  
  def testNoneNodeIncluded(self): # -> None:
    ...
  
  def testCustomClassMapWithPath(self): # -> None:
    class ExampleClass(Mapping[Any, Any]):
      """Small example custom class."""
      ...
    
    
  


if __name__ == "__main__":
  ...



"""
Episode Functions Utils
========================

Work in progress...
"""

from typing import Dict, Tuple, Any, List, Optional
import os
import numpy as np
from numpy.typing import NDArray
from numpy import float64
import json
import pandas as pd
import datetime
from eprllib import logger

def load_ep_model(model_path: str) -> Dict[str, Any]:
    """
    Load the episode model from the given path.

    Args:
        model_path (str): The path to the episode model.

    Returns:
        Dict: The loaded episode model.
    """
    
    
    # Check that the file exists and finish with .epJSON
    if not model_path.endswith('.epJSON'):
        msg = f'EpisodeFunctionUtils: The file {model_path} is not a .epJSON file'
        logger.error(msg)
        raise ValueError(msg)
    
    # Open the file
    with open(model_path, 'rb') as f:
        model: Dict[str, Any] = json.load(f)
    return model

def save_ep_model(model: Dict[str, Any], model_folder_path: str) -> str:
    """
    Save the episode model to the given path.

    Args:
        model (Dict): The episode model to save.
        model_folder_path (str): The path of the folder where to save the episode model.

    Returns:
        str: The path to the saved episode model.
    """
    model_path = f"{model_folder_path}/{os.getpid():05}.epJSON"
    with open(model_path, 'w') as fp:
        json.dump(model, fp, sort_keys=False, indent=4)
        # The new modify epjson file is writed.
    return model_path

def get_random_weather(epw_files_folder_path:str) -> str:
    """
    This method select a random epw file from the folder and return the path.

    Args:
        epw_files_folder_path (str): Folder that contain the epw files.

    Return:
        str: The path to the epw file selected
    """
    # Check that the folder has at least one file of type .epw
    if not any(file.endswith('.epw') for file in os.listdir(epw_files_folder_path)):
        msg = "EpisodeFunctionUtils: The folder does not contain any .epw files."
        logger.error(msg)
        raise ValueError(msg)
    
    # Build a list of the files that have the extension .epw
    epw_files = [file for file in os.listdir(epw_files_folder_path) if file.endswith('.epw')]
    
    # select a random epw file that has the extension .epw
    id_epw_file = np.random.randint(0, len(epw_files))
    
    # The path to the epjson file is defined and returned
    return os.path.join(epw_files_folder_path, epw_files[id_epw_file])

def run_period(julian_day:int, days_period:int=28) -> Tuple[int,int,int,int]:
    """
    This function returns the begin and end date of the 'days_period' of simulation given a 'julian_day' between 1 and 365.
    """
    if julian_day < 1 or julian_day+days_period > 365:
        msg = 'EpisodeFunctionUtils: Julian day must be between 1 and (365-days_period)'
        logger.error(msg)
        raise ValueError(msg)
    if days_period < 0:
        msg = 'EpisodeFunctionUtils: Days period must be integer and positive'
        logger.error(msg)
        raise ValueError(msg)
    
    # Declaration of variables
    beging_month = 1
    beging_day = 0
    check_day = 0
    max_day = max_day_in_month(beging_month)
    
    # Initial date
    while True:
        beging_day += 1
        check_day += 1
        if julian_day == check_day:
            break
        if beging_day >= max_day:
            beging_day = 0
            beging_month += 1
            max_day = max_day_in_month(beging_month)
    
    # End date
    end_month = beging_month
    end_day = beging_day + days_period
    while True:
        if end_day > max_day:
            end_day -= max_day
            end_month += 1
            max_day = max_day_in_month(end_month)
        else:
            break
        
    return beging_month, beging_day, end_month, end_day


def max_day_in_month(month:int) -> int:
    """
    This function returns the maximum number of days in a given month.
    """
    months_31 = [1,3,5,7,8,10,12]
    months_30 = [4,6,9,11]
    months_28 = [2]
    if month in months_31:
        max_day = 31
    elif month in months_30:
        max_day = 30
    elif month in months_28:
        max_day = 28
    else:
        msg = 'EpisodeFunctionUtils: Invalid month'
        logger.error(msg)
        raise ValueError(msg)
    return max_day

def building_dimension(
    epJSON_object: Dict[str, Any], 
    h:float, 
    w:float, 
    l:float, 
    window_area_relation: NDArray[float64],
    ) -> Dict[str, Any]:
    """
    This function modify the building dimensions and the windows positions in the epJSON file.

    Args:
        epJSON_object (Dict): The epJSON file.
        h (float): The height of the building.
        w (float): The width of the building.
        l (float): The length of the building.
        window_area_relation (NDarray): The proportion of the windows in the building. The order is [north, east, south, west].

    Returns:
        Dict: The modified epJSON file.
    """
    # Reshape the position vectors
    surface_coordenates: Dict[str, Any] = {
        'wall_north': [[0,0,h],[0,0,0],[w,0,0],[w,0,h]],
        'wall_east': [[w,0,h],[w,0,0],[w,l,0],[w,l,h]],
        'wall_south': [[w,l,h],[w,l,0],[0,l,0],[0,l,h]],
        'wall_west': [[0,l,h],[0,l,0],[0,0,0],[0,0,h]],
        'roof': [[w,0,h],[w,l,h],[0,l,h],[0,0,h]],
        'floor': [[w,l,0],[w,0,0],[0,0,0],[0,l,0]]
    }
    coordinate = ['x', 'y', 'z']
    for surface_name in [key for key in surface_coordenates.keys()]:
        # Iterate over the four vertices of the surface
        for _ in range(4):
            for xyz in range(3):
                epJSON_object['BuildingSurface:Detailed'][surface_name]['vertices'][_]['vertex_'+coordinate[xyz]+'_coordinate'] = surface_coordenates[surface_name][_][xyz]
    
    # Calculate the centroid of each wall
    centroid_coordenates: Dict[str, Any] = {
        'wall_north': [w/2,0,h/2],
        'wall_east': [w,l/2,h/2],
        'wall_south': [w/2,l,h/2],
        'wall_west': [0,l/2,h/2],
        'roof': [w/2,l/2,h],
        'floor': [w/2,l/2,0]
    }
    
    # Calculate the relative coordinates of each surface vertex with respect to its centroid
    relative_coordenates: Dict[str, Any] = {}
    for surface_name, vertices in surface_coordenates.items():
        centroid = np.array(centroid_coordenates[surface_name])
        relative_vertices = [np.array(vertex) - centroid for vertex in vertices]
        relative_coordenates[surface_name] = relative_vertices
    
    window_area_relation_dict: Dict[str, float] = {
        'wall_north': float(window_area_relation[0]),
        'wall_east': float(window_area_relation[1]),
        'wall_south': float(window_area_relation[2]),
        'wall_west': float(window_area_relation[3])
    }
    
    # Calculate the relative coordinates of the windows
    relative_window_coordenates: Dict[str, Any] = {}

    wall_dimensions: Dict[str, Dict[str, float]] = {
        'wall_north': {'width': w, 'height': h},
        'wall_east': {'width': l, 'height': h},
        'wall_south': {'width': w, 'height': h},
        'wall_west': {'width': l, 'height': h}
    }


    for surface_name, relation in window_area_relation_dict.items():
        if relation > 0 and surface_name in wall_dimensions:  # Only consider walls with windows
            centroid = np.array(centroid_coordenates[surface_name])
            wall_width = wall_dimensions[surface_name]['width']
            wall_height = wall_dimensions[surface_name]['height']

            # Calculate window dimensions assuming the window is centered and scales proportionally
            window_area = wall_width * wall_height * relation
            # Assuming the window maintains the same aspect ratio as the wall for simplicity
            # We can adjust this if a specific aspect ratio is needed
            window_width = np.sqrt(window_area * wall_width / wall_height)
            window_height = np.sqrt(window_area * wall_height / wall_width)


            # Calculate window coordinates relative to the wall centroid
            # Assuming the centroid is at the center of the wall
            half_window_width = window_width / 2
            half_window_height = window_height / 2

            if surface_name == 'wall_north':
                window_coords = [
                    [centroid[0] - half_window_width, centroid[1], centroid[2] + half_window_height],
                    [centroid[0] - half_window_width, centroid[1], centroid[2] - half_window_height],
                    [centroid[0] + half_window_width, centroid[1], centroid[2] - half_window_height],
                    [centroid[0] + half_window_width, centroid[1], centroid[2] + half_window_height],
                ]
            elif surface_name == 'wall_east':
                window_coords = [
                    [centroid[0], centroid[1] - half_window_width, centroid[2] + half_window_height],
                    [centroid[0], centroid[1] - half_window_width, centroid[2] - half_window_height],
                    [centroid[0], centroid[1] + half_window_width, centroid[2] - half_window_height],
                    [centroid[0], centroid[1] + half_window_width, centroid[2] + half_window_height],
                ]
            elif surface_name == 'wall_south':
                window_coords = [
                    [centroid[0] + half_window_width, centroid[1], centroid[2] + half_window_height],
                    [centroid[0] + half_window_width, centroid[1], centroid[2] - half_window_height],
                    [centroid[0] - half_window_width, centroid[1], centroid[2] - half_window_height],
                    [centroid[0] - half_window_width, centroid[1], centroid[2] + half_window_height],
                ]
            elif surface_name == 'wall_west':
                window_coords = [
                    [centroid[0], centroid[1] + half_window_width, centroid[2] + half_window_height],
                    [centroid[0], centroid[1] + half_window_width, centroid[2] - half_window_height],
                    [centroid[0], centroid[1] - half_window_width, centroid[2] - half_window_height],
                    [centroid[0], centroid[1] - half_window_width, centroid[2] + half_window_height],
                ]
            else:
                continue


            relative_window_vertices = [np.array(vertex) - centroid for vertex in window_coords]
            relative_window_coordenates[f'window_{surface_name.split("_")[1]}'] = relative_window_vertices

    # Transform relative window coordinates back to original coordinates
    window_coordenates_original: Dict[str, Any] = {}

    for window_name, relative_vertices in relative_window_coordenates.items():
        # Extract the original wall surface name from the window name
        surface_name = f"wall_{window_name.split('_')[1]}"
        centroid = np.array(centroid_coordenates[surface_name])
        original_vertices = [vertex + centroid for vertex in relative_vertices]
        window_coordenates_original[window_name] = original_vertices
    
    # Change the values in the epJSON file
    for window_name in [key for key in window_coordenates_original.keys()]:
        # Iterate over the four vertices of the surface
        for _ in range(1,5,1):
            for xyz in range(3):
                epJSON_object['FenestrationSurface:Detailed'][window_name]['vertex_'+str(_)+'_'+coordinate[xyz]+'_coordinate'] = window_coordenates_original[window_name][_-1][xyz]
        
    return epJSON_object


def window_size_epJSON(epJSON_object: Dict[str, Any], window:str, area_ventana:float) -> Dict[str, Any]:
    """
    Given a epJSON_object, return another epJSON_object with diferent size of windows.

    Args:
        epJSON_object (json): _description_
        window_name (str): _description_
        factor (float): _description_

    Returns:
        json: Devuelve el objeto epJSON modificado.
    """
    window_vertexs: Dict[str, Any] = {}
    # se extraen los valores de los vértices de cada ventana según el epJSON
    window_vertexs['v1x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_1_x_coordinate']
    window_vertexs['v1y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_1_y_coordinate']
    window_vertexs['v1z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_1_z_coordinate']
    window_vertexs['v2x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_2_x_coordinate']
    window_vertexs['v2y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_2_y_coordinate']
    window_vertexs['v2z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_2_z_coordinate']
    window_vertexs['v3x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_3_x_coordinate']
    window_vertexs['v3y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_3_y_coordinate']
    window_vertexs['v3z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_3_z_coordinate']
    window_vertexs['v4x'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_4_x_coordinate']
    window_vertexs['v4y'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_4_y_coordinate']
    window_vertexs['v4z'] = epJSON_object['FenestrationSurface:Detailed'][window]['vertex_4_z_coordinate']
    
    L: List[List[float]] = [] # agrupa los vertices en forma de lista: [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3],[x4,y4,z4]]
    for l in range(1,5,1):
        vertex_x = 'v'+str(l)+'x'
        vertex_y = 'v'+str(l)+'y'
        vertex_z = 'v'+str(l)+'z'
        L.append([window_vertexs[vertex_x], window_vertexs[vertex_y], window_vertexs[vertex_z]])

    # Se calcula el factor de escala de la ventana
    area_ventana_old: float = fenestration_area(epJSON_object, window)
    factor_escala: float = area_ventana/area_ventana_old
    centro: List[float] = calcular_centro(L)
    ventana_escalada: List[List[float]] = []
    for punto in L:
        nuevo_punto = [centro[0] + (punto[0] - centro[0]) * factor_escala**(1/2),
                    centro[1] + (punto[1] - centro[1]) * factor_escala**(1/2),
                    centro[2] + (punto[2] - centro[2]) * factor_escala**(1/2)]
        ventana_escalada.append(nuevo_punto)
    
    for l in range(1,5,1):
        epJSON_object['FenestrationSurface:Detailed'][window]['vertex_'+str(l)+'_x_coordinate'] = ventana_escalada[l-1][0]
        epJSON_object['FenestrationSurface:Detailed'][window]['vertex_'+str(l)+'_y_coordinate'] = ventana_escalada[l-1][1]
        epJSON_object['FenestrationSurface:Detailed'][window]['vertex_'+str(l)+'_z_coordinate'] = ventana_escalada[l-1][2]

    return epJSON_object

def calcular_centro(surface: List[List[float]]) -> List[float]:
    # Calcula el centro de un cuadrilátero
    return [(surface[0][0] + surface[1][0] + surface[2][0] + surface[3][0]) / 4,
            (surface[0][1] + surface[1][1] + surface[2][1] + surface[3][1]) / 4,
            (surface[0][2] + surface[1][2] + surface[2][2] + surface[3][2]) / 4]


def generate_variated_schedule(
    input_path: str,
    output_path: str,
    variation_percentage: float = 0.1):
    """
    Generates a new CSV with variations in schedule data.
    
    Parameters:
    - input_path: str, path to the input CSV file.
    - output_path: str, path to save the output CSV file.
    - variation_percentage: float, maximum variation for each value (default: 0.1).
    """
    # Load the input CSV
    assert os.path.exists(input_path), f"Input file {input_path} does not exist."
    assert input_path.endswith('.csv'), "Input file must be a CSV."
    
    data = pd.read_csv(input_path)
    
    # Calculate variation range
    variation_factor = variation_percentage
    
    # Apply random variations within the specified range
    def apply_variation(value: float) -> float:
        variation = np.random.uniform(-variation_factor, variation_factor)
        new_value = value * (1 + variation)
        return np.clip(new_value, 0, 1)  # Ensure values stay within [0, 1]

    # Apply variation to all columns
    variated_data: pd.DataFrame = data.applymap(apply_variation)
    
    # Save the new dataset to the specified output path
    variated_data.to_csv(output_path, index=False)

def generate_variated_schedule_with_shift(
    input_path: str, 
    output_path: str, 
    variation_percentage: float = 10, 
    shift_hours: int = 0
    ) -> None:
    """
    Generates a new CSV with variations and optional time shift in schedule data.
    
    Parameters:
    - input_path: str, path to the input CSV file.
    - output_path: str, path to save the output CSV file.
    - variation_percentage: float, maximum percentage variation for each value (default: 10).
    - shift_hours: int, number of hours to shift the schedules (default: 0).
    """
    # Load the input CSV
    data = pd.read_csv(input_path)
    
    # Calculate variation range
    variation_factor = variation_percentage / 100.0
    
    # Apply random variations within the specified range
    def apply_variation(value: float):
        variation = np.random.uniform(-variation_factor, variation_factor)
        new_value = value * (1 + variation)
        return np.clip(new_value, 0, 1)  # Ensure values stay within [0, 1]
    
    # Apply variation to all columns
    variated_data = data.applymap(apply_variation)
    
    # Apply temporal shift if specified
    if shift_hours != 0:
        rows_to_shift = shift_hours  # Number of rows to shift (1 hour per row)
        variated_data = variated_data.shift(periods=rows_to_shift, fill_value=0).reset_index(drop=True)

    # Save the new dataset to the specified output path
    variated_data.to_csv(output_path, index=False)

def generate_occupancy_schedule(
    input_path: str, 
    output_path: str, 
    random_variation: bool = False, 
    seed: Optional[int] = None
    ) -> None:
    """
    Generates an updated occupancy schedule based on typical residential patterns.
    
    Parameters:
    - input_path: str, path to the input CSV file (with "occupancy" column).
    - output_path: str, path to save the output CSV file.
    - random_variation: bool, whether to introduce random variations to the schedule.
    - seed: int, random seed for reproducibility (default: None).
    """
    # Set random seed if needed
    if seed is not None:
        np.random.seed(seed)
    
    # Load the input CSV
    data = pd.read_csv(input_path)
    assert "occupancy" in data.columns, "Input file must have an 'occupancy' column."
    
    # Generate timestamps assuming hourly data for a year (if not provided)
    num_rows = len(data)
    start_date = datetime.datetime(2024, 1, 1)
    timestamps = [start_date + datetime.timedelta(hours=i) for i in range(num_rows)]
    data["timestamp"] = timestamps

    # Define occupancy rules based on weekdays and hours
    def occupancy_pattern(timestamp: datetime.datetime) -> int:
        hour = timestamp.hour
        day = timestamp.weekday()  # 0=Monday, 6=Sunday
        
        if day < 5:  # Weekdays
            if 7 <= hour < 9 or 17 <= hour < 23:
                return 1  # Morning/evening peak
            elif 9 <= hour < 17:
                return 0  # Work hours
            else:
                return 1  # Night hours (sleeping)
        else:  # Weekends
            if 9 <= hour < 23:
                return 1  # Occupied during the day
            else:
                return 1  # Night hours (sleeping)

    # Apply patterns
    data["occupancy"] = data["timestamp"].apply(occupancy_pattern)
    
    # Introduce random variations if enabled
    if random_variation:
        def add_randomness(occ_value):
            return occ_value if np.random.rand() > 0.1 else 1 - occ_value  # Flip value with 10% probability
        
        data["occupancy"] = data["occupancy"].apply(add_randomness)

    # Drop the timestamp column before saving
    data.drop(columns=["timestamp"], inplace=True)

    # Save the updated dataset to the specified output path
    data.to_csv(output_path, index=False)
    

def inertial_mass_calculation(env_config:Dict[str, Any]) -> float:
    """
    The function reads the epjson file path from the env_config dictionary. If the epjson
    key is not present, it raises a ValueError. Otherwise, it opens the specified epjson
    file, loads its contents as a Python dictionary (epJSON_object), and then calls the 
    inertial_mass function (not shown in the provided code) with epJSON_object
    as an argument. The result of inertial_mass is returned as a float value.

    Args:
        env_config (dict): A dictionary containing the environment configuration, including the 
            episode_config key, which should have an epjson key with the path to the 
            EnergyPlus JSON file.

    Returns:
        float: The inertial mass calculated based on the contents of the epjson file.
    
    Example:
        env_config = {
            'episode_config': {
                'epjson': 'path/to/epjson_file.json'
            }
        }

        inertial_mass_value = inertial_mass_calculation(env_config)
        print(f"Inertial mass: {inertial_mass_value}")

    """
    # If the path to epjson is not set, arraise a error.
    if not env_config.get('epjson_path', False):
        msg = 'EpisodeFunctionUtils: epjson_path is not defined in env_config'
        logger.error(msg)
        raise ValueError(msg)
    
    with open(env_config['epjson_path']) as file:
        epJSON_object: dict = json.load(file)

    return effective_thermal_capacity(epJSON_object)

def effective_thermal_capacity(epJSON_object:Dict[str,Any]) -> float:
    """
    The `effective_thermal_capacity` function calculates the total effective thermal capacity of a building based 
    on the construction materials and surface areas specified in an EnergyPlus JSON 
    object.

    Args:
        epJSON_object (dict[str,dict]): A dictionary containing information about the 
        building surfaces, materials, and constructions in the EnergyPlus JSON format. 
        The keys of the outer dictionary represent different object types (e.g., 
        "BuildingSurface:Detailed", "Material", "Construction"), and the values are 
        dictionaries containing the properties of each object instance.

    Returns:
        float: The total thermal capacity of the building in J/°C (Joules per degree Celsius). 
        The thermal mass represents the amount of heat energy required to raise the 
        temperature of the building's construction materials by one degree Celsius.
    """
    # se define una lista para almacenar
    Cdyn = []
    
    building_surfaces = []
    if epJSON_object.get("BuildingSurface:Detailed", False):
        building_surfaces = [key for key in epJSON_object["BuildingSurface:Detailed"].keys()]
    # se obtienen los nombres de las superficies de la envolvente
    internal_mass_surfaces = []
    if epJSON_object.get("InternalMass", False):
        internal_mass_surfaces = [key for key in epJSON_object["InternalMass"].keys()]
    
    all_building_keys = [key for key in epJSON_object.keys()]
    all_material_list = ['Material','Material:NoMass','Material:InfraredTransparent','Material:AirGap',
        'Material:RoofVegetation','WindowMaterial:SimpleGlazingSystem','WindowMaterial:Glazing',
        'WindowMaterial:GlazingGroup:Thermochromic','WindowMaterial:Glazing:RefractionExtinctionMethod',
        'WindowMaterial:Gas','WindowGap:SupportPillar','WindowGap:DeflectionState',
        'WindowMaterial:GasMixture','WindowMaterial:Gap'
    ]
    materials_dict = {}
    for material in all_material_list:
        if material in all_building_keys:
            materials_dict[material] = epJSON_object[material].keys()
    # se obtienen los nombres de los diferentes tipos de materiales
    
    # lazo para consultar cada superficie de la envolvente
    for surface in building_surfaces:
        # se calcula el área de la superficie
        area = material_area(epJSON_object,surface)
        # se identifica la consutrucción
        s_construction = epJSON_object['BuildingSurface:Detailed'][surface]['construction_name']
        
        # se obtiene la densidad del materia: \rho[kg/m3]
        # se obtiene el calor específico del material: C[J/kg°C]
        # se calcula el volumen que ocupa el material: V[m3]=area*thickness
        # se calcula la capacidad térmica: M[J/°C] = \rho[kg/m3] * C[J/kg°C] * V[m3]
        
        # se establece un lazo para calcular la masa térmica de cada capa
        Cdyn_surface = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap' or material_list == 'Material:InfraredTransparent' or material_list == 'WindowMaterial:Gas':
                Cdyn_capa = 0
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                calor_especifico_capa = epJSON_object[material_list][material]['specific_heat']
                densidad_capa = epJSON_object[material_list][material]['density']
                Cdyn_capa = (area * espesor_capa * densidad_capa) * calor_especifico_capa

            # se suma la resistencia de la superficie
            Cdyn_surface += Cdyn_capa
        # se guarda la resistencia de la superficie
        Cdyn.append(Cdyn_surface)
    
    # se suma la masa interna asignada
    for surface in internal_mass_surfaces:
        # se calcula el área de la superficie
        area = epJSON_object['InternalMass'][surface]['surface_area']
        # se identifica la consutrucción
        s_construction = epJSON_object['InternalMass'][surface]['construction_name']
        
        # se obtiene la densidad del materia: \rho[kg/m3]
        # se obtiene el calor específico del material: C[J/kg°C]
        # se calcula el volumen que ocupa el material: V[m3]=area*thickness
        # se calcula la capacidad térmica: M[J/°C] = \rho[kg/m3] * C[J/kg°C] * V[m3]
        
        # se establece un lazo para calcular la masa térmica de cada capa
        Cdyn_surface = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap' or material_list == 'Material:InfraredTransparent' or material_list == 'WindowMaterial:Gas':
                Cdyn_capa = 0
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                calor_especifico_capa = epJSON_object[material_list][material]['specific_heat']
                densidad_capa = epJSON_object[material_list][material]['density']
                Cdyn_capa = (area * espesor_capa * densidad_capa) * calor_especifico_capa

            # se suma la resistencia de la superficie
            Cdyn_surface += Cdyn_capa
        # se guarda la resistencia de la superficie
        Cdyn.append(Cdyn_surface)
    
    # Cálculo de la masa termica total
    Cdyn_total = 0.
    for m in range(0,len(Cdyn)-1,1):
        Cdyn_total += Cdyn[m]
    
    return Cdyn_total

def u_factor_calculation(env_config:Dict[str,Any]) -> float:
    """
    The u_factor_calculation function calculates the U-factor (a measure of heat 
    transfer through a building element) based on the provided environment configuration 
    (env_config).

    Args:
        env_config (dict): A dictionary containing the environment configuration, including 
        the episode_config key, which should have an epjson key with the path to the 
        EnergyPlus JSON file.

    Returns:
        float: The calculated U-factor value.
        
    Example:
        env_config = {
            'episode_config': {
                'epjson': 'path/to/epjson_file.json'
            }
        }

        u_factor_value = u_factor_calculation(env_config)
        print(f"U-factor: {u_factor_value}")

    """
    # If the path to epjson is not set, arraise a error.
    if not env_config.get('epjson_path', False):
        msg = 'EpisodeFunctionUtils: epjson_path is not defined in env_config'
        logger.error(msg)
        raise ValueError(msg)
    with open(env_config['epjson_path']) as file:
        epJSON_object: dict = json.load(file)
    #  Calculate the u_factor
    return u_factor(epJSON_object)

def u_factor(epJSON_object:Dict[str,Any]) -> float:
    """
    This function select all the building surfaces and fenestration surfaces and calculate the
    global U-factor of the building, like EnergyPlus does.
    """
    # se define una lista para almacenar las resistencias de cada supercie
    resistences = []
    areas = []
    # se obtienen los nombres de las superficies de la envolvente
    building_surfaces = [key for key in epJSON_object['BuildingSurface:Detailed'].keys()]
    fenestration_surfaces = [key for key in epJSON_object['FenestrationSurface:Detailed'].keys()]
    # se obtienen los nombres de los diferentes tipos de materiales
    
    all_building_keys = [key for key in epJSON_object.keys()]
    all_material_list = ['Material','Material:NoMass','Material:InfraredTransparent','Material:AirGap',
        'Material:RoofVegetation','WindowMaterial:SimpleGlazingSystem','WindowMaterial:Glazing',
        'WindowMaterial:GlazingGroup:Thermochromic','WindowMaterial:Glazing:RefractionExtinctionMethod',
        'WindowMaterial:Gas','WindowGap:SupportPillar','WindowGap:DeflectionState',
        'WindowMaterial:GasMixture','WindowMaterial:Gap'
    ]
    materials_dict = {}
    for material in all_material_list:
        if material in all_building_keys:
            materials_dict[material] = epJSON_object[material].keys()
    # se obtienen los nombres de los diferentes tipos de materiales
    
    # lazo para consultar cada superficie de la envolvente
    for surface in building_surfaces:
        # se calcula el área de la superficie
        areas.append(material_area(epJSON_object,surface))
        # se identifica la consutrucción
        s_construction = epJSON_object['BuildingSurface:Detailed'][surface]['construction_name']
        # se establece un lazo para calcular la resistencia de cada capa
        r_surface = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap':
                r_capa = epJSON_object[material_list][material]['thermal_resistance']
            elif material_list == 'Material:InfraredTransparent':
                r_capa = 0
            elif material_list == 'WindowMaterial:Gas':
                espesor_capa = epJSON_object[material_list][material]['thickness']
                if epJSON_object[material_list][material]['gas_type'] == 'Air':
                    conductividad_capa = 0.0257
                elif epJSON_object[material_list][material]['gas_type'] == 'Argon':
                    conductividad_capa = 0.0162
                elif epJSON_object[material_list][material]['gas_type'] == 'Xenon':
                    conductividad_capa = 0.00576
                elif epJSON_object[material_list][material]['gas_type'] == 'Krypton':
                    conductividad_capa = 0.00943
                else:
                    msg =('EpisodeFunctionUtils: El nombre del gas no corresponde con los que pueden utilizarse: Air, Argon, Xenon, Krypton.')
                    logger.error(msg)
                    raise KeyError(msg)
                r_capa = espesor_capa/conductividad_capa
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                conductividad_capa = epJSON_object[material_list][material]['conductivity']
                r_capa = espesor_capa/conductividad_capa

            # se suma la resistencia de la superficie
            r_surface += r_capa
        # se guarda la resistencia de la superficie
        resistences.append(r_surface)
        
    # lazo para consultar cada superfice de fenestración
    for fenestration in fenestration_surfaces:
        # se calcula el área de la superficie
        areas.append(fenestration_area(epJSON_object, fenestration))
        # se identifica la consutrucción
        s_construction = epJSON_object['FenestrationSurface:Detailed'][fenestration]['construction_name']
        # se establece un lazo para calcular la resistencia de cada capa
        r_surface = 0
        u_factor_windows = 0
        u_factor_window = 0
        layers = [key for key in epJSON_object['Construction'][s_construction].keys()]
        for layer in layers:
            material = epJSON_object['Construction'][s_construction][layer]
            material_list = find_dict_key_by_nested_key(
                material,
                materials_dict
            )
            # se obtiene el espesor y la conductividad térmica del material de la capa
            if material_list == 'Material:NoMass' or material_list == 'Material:AirGap':
                r_capa = epJSON_object[material_list][material]['thermal_resistance']
            elif material_list == 'Material:InfraredTransparent':
                r_capa = 0
            elif material_list == 'WindowMaterial:Gas':
                espesor_capa = epJSON_object[material_list][material]['thickness']
                if epJSON_object[material_list][material]['gas_type'] == 'Air':
                    conductividad_capa = 0.0257
                elif epJSON_object[material_list][material]['gas_type'] == 'Argon':
                    conductividad_capa = 0.0162
                elif epJSON_object[material_list][material]['gas_type'] == 'Xenon':
                    conductividad_capa = 0.00576
                elif epJSON_object[material_list][material]['gas_type'] == 'Krypton':
                    conductividad_capa = 0.00943
                else:
                    print('El nombre del gas no corresponde con los que pueden utilizarse: Air, Argon, Xenon, Krypton.')
                    NameError
                r_capa = espesor_capa/conductividad_capa
            elif material_list == 'WindowMaterial:SimpleGlazingSystem':
                u_factor_window = epJSON_object[material_list][material]['u_factor']
            else:
                espesor_capa = epJSON_object[material_list][material]['thickness']
                conductividad_capa = epJSON_object[material_list][material]['conductivity']
                r_capa = espesor_capa/conductividad_capa

            # se suma la resistencia de la superficie
            r_surface += r_capa
            u_factor_windows += u_factor_window
        # se guarda la resistencia de la superficie
        resistences.append(r_surface)

    # Cálculo de U-Factor en W/°C
    u_factor = 0.
    for n in range(0, len(areas)-1,1):
        u_factor =+ areas[n]/resistences[n]
    
    return u_factor + u_factor_windows

def material_area(epJSON_object: Dict[str, Any],nombre_superficie: str):
    """
    _summary_

    Args:
        epJSON_object (_type_): _description_
        nombre_superficie (_type_): _description_

    Returns:
        _type_: _description_
    """
    # Calcula dos vectores que forman dos lados del cuadrilátero
    vector1 = [
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][1]['vertex_x_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_x_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][1]['vertex_y_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_y_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][1]['vertex_z_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_z_coordinate']
    ]
    vector2 = [
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][2]['vertex_x_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_x_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][2]['vertex_y_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_y_coordinate'],
        epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][2]['vertex_z_coordinate'] - epJSON_object['BuildingSurface:Detailed'][nombre_superficie]['vertices'][0]['vertex_z_coordinate']
    ]

    # Calcula el producto vectorial entre los dos vectores
    producto_vectorial = [
        vector1[1] * vector2[2] - vector1[2] * vector2[1],
        vector1[2] * vector2[0] - vector1[0] * vector2[2],
        vector1[0] * vector2[1] - vector1[1] * vector2[0]
    ]

    # Calcula el módulo del producto vectorial como el área del cuadrilátero
    area = 0.5 * (abs(producto_vectorial[0]) + abs(producto_vectorial[1]) + abs(producto_vectorial[2]))
    return area

def fenestration_area(epJSON_object: Dict[str, Any], fenestration: str) -> float:
    """
    _summary_

    Args:
        epJSON_object (_type_): _description_
        fenestration (_type_): _description_

    Returns:
        _type_: _description_
    """
    
    # Calcula dos vectores que forman dos lados del cuadrilátero
    vector1 = [
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_2_x_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_x_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_2_y_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_y_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_2_z_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_z_coordinate']
    ]
    vector2 = [
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_3_x_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_x_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_3_y_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_y_coordinate'],
        epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_3_z_coordinate'] - epJSON_object['FenestrationSurface:Detailed'][fenestration]['vertex_1_z_coordinate']
    ]

    # Calcula el producto vectorial entre los dos vectores
    producto_vectorial = [
        vector1[1] * vector2[2] - vector1[2] * vector2[1],
        vector1[2] * vector2[0] - vector1[0] * vector2[2],
        vector1[0] * vector2[1] - vector1[1] * vector2[0]
    ]

    # Calcula el módulo del producto vectorial como el área del cuadrilátero
    area = 0.5 * (abs(producto_vectorial[0]) + abs(producto_vectorial[1]) + abs(producto_vectorial[2]))
    return area

def find_dict_key_by_nested_key(key: str, lists_dict: Dict[str, List[str]]):
    """
    _summary_

    Args:
        key (_type_): _description_
        lists_dict (_type_): _description_

    Returns:
        _type_: _description_
    """
    for dict_key, lst in lists_dict.items():
        if key in lst:
            return dict_key
    return None


def run_period_change(
    env_config:Dict[str,Any],
    init_julian_day: Optional[int|str] = None,
    end_julian_day: Optional[int] = None,
    init_month: Optional[int] = None,
    init_day: Optional[int] = None,
    end_month: Optional[int] = None,
    end_day: Optional[int] = None,
    simulation_duration: Optional[int] = None
    ) -> Dict[str, Any]:
    """
    This function is used to modify the RunPeriod longitude of a epJSON file.
    
    Args:
        epjson_file(str): path to the epJSON file.
        output_folder(str): path to the destination folder where the modified file will be saved.
        episode_len(int)[Optional]: longitude of the RunPeriod, or episode in the context of eprllib. Default is 7.
        init_julian_day(int): The initial julian day to determine the RunPeriod. Defaut is 0, that means a random choice.
        
    Return:
        str: path to the modified epJSON file.
    """
    env_configuration = env_config
    
    # Check that (init_julian_day or (init_month and init_day)) and (end_julian_day or (end_month and end_day)) are not both None
    if (init_julian_day is None and (init_month is None or init_day is None)) or (end_julian_day is None and (end_month is None or end_day is None)):
        msg = "EpisodeFunctionUtils: Both init_julian_day or (init_month and init_day) and end_julian_day or (end_month and end_day) must be provided."
        logger.error(msg)
        raise ValueError(msg)
    # check that if end_julian_day is None and (end_month is None or end_day is None) -> simulation_duration must be provided
    if end_julian_day is None and (end_month is None or end_day is None) and simulation_duration is None:
        msg = "EpisodeFunctionUtils: If end_julian_day is None and (end_month is None or end_day is None), simulation_duration must be provided."
        logger.error(msg)
        raise ValueError(msg)
    # check that if simulation_duration is provided, it must be an integer and lower that 364-init_julian_day
    assert isinstance(simulation_duration, int), "EpisodeFunctionUtils: simulation_duration must be an integer."
    
    # Open the epjson file
    with open(env_configuration['epjson_path']) as epf:
        epjson_object = pd.read_json(epf)
        
    # Transform the julian day into day,month tuple
    if init_julian_day is not None:
        if init_julian_day == 'random':
            init_julian_day = np.random.randint(1, 366-simulation_duration)
            init_day, init_month = from_julian_day(init_julian_day)
            end_day, end_month = from_julian_day(init_julian_day + simulation_duration)
            
        elif isinstance(init_julian_day, int):
            init_day, init_month = from_julian_day(init_julian_day)
            if end_julian_day != None:
                end_day, end_month = from_julian_day(end_julian_day)
            else:
                # check that end_day and end_month are different that None
                if end_day is None or end_month is None:
                    msg = "EpisodeFunctionUtils: If end_julian_day is None, simulation_duration must be provided."
                    logger.error(msg)
                    raise ValueError(msg)
                
        else:
            msg = "EpisodeFunctionUtils: init_julian_day must be an integer or 'random'."
            logger.error(msg)
            raise ValueError(msg)
    
    assert isinstance(init_julian_day, int), "EpisodeFunctionUtils: init_julian_day must be an integer."
    assert simulation_duration < 364-init_julian_day, "EpisodeFunctionUtils: the duration of the episode cannot exceds the year."
      
    # Change the values in the epjson file
    epjson_object['RunPeriod']['RunPeriod 1']['beging_month'] = init_month
    epjson_object['RunPeriod']['RunPeriod 1']['begin_day_of_month'] = init_day
    epjson_object['RunPeriod']['RunPeriod 1']['end_month'] = end_month
    epjson_object['RunPeriod']['RunPeriod 1']['end_day_of_month'] = end_day
    # Save the epjson file modified into the output folder
    df = pd.DataFrame(epjson_object)
    env_configuration['epjson_path'] = env_configuration['output_path'] + f'/epjson_file_{init_julian_day}.epjson'
    df.to_json(env_configuration['epjson_path'], orient='records')

    return env_configuration

def from_julian_day(julian_day:int) -> Tuple[int,int]:
    """
    This funtion take a julian day and return the corresponding
    day and month for a tipical year of 365 days.
    
    Args:
        julian_day(int): Julian day to be transform
        
    Return:
        Tuple[int,int]: (day,month)
    """
    # Define the number of days in each month
    days_in_months = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    # Define the day variable as equal to julian day and discount it
    day = julian_day
    for month, days_in_month in enumerate(days_in_months):
        if day <= days_in_month:
            return (day, month + 1)
        day -= days_in_month
    
    return (day, 12)
        

def extract_epw_location_data(epw_file_path: str) -> Tuple[float, float, float, float]:
    """
    Extrae la latitud, longitud, huso horario y altitud de un archivo EPW.

    Args:
        epw_file_path (str): La ruta al archivo de clima en formato EPW.

    Returns:
        tuple: Una tupla que contiene (latitud, longitud, huso horario, altitud).
               Retorna None si el archivo no es válido o no se encuentra.
    """
    try:
        # Lee el archivo EPW. Los primeros 8-9 son la información del encabezado
        # dependiendo de la versión del archivo.
        # La información de la ubicación está en la segunda línea.
        header_lines: List[str] = []
        with open(epw_file_path, 'r', encoding='utf-8') as f:
            for _ in range(8):
                try:
                    header_lines.append(next(f))
                except StopIteration:
                    break # Stop if we reach the end of the file

        # Check if we have at least one line to process for location
        if not header_lines:
            print(f"Error: El archivo EPW está vacío o no contiene líneas de encabezado: {epw_file_path}")
            return 0.,0.,0.,0.

        # El encabezado EPW tiene campos separados por comas.
        # La línea de ubicación es la primera (índice 0).
        location_line: List[str] = header_lines[0].strip().split(',')

        # Los índices corresponden a los campos en la línea de ubicación EPW:
        # [0] Ciudad, [1] Estado/Provincia, [2] País/Región,
        # [3] Fuente, [4] WMO, [5] Latitud,
        # [6] Longitud, [7] Huso Horario, [8] Altitud

        # Ensure we have enough fields in the location line
        if len(location_line) < 9:
             print(f"Error: La línea de ubicación en el archivo EPW no tiene el formato esperado. Se encontraron {len(location_line)} campos, se esperaban al menos 9.")
             return 0.,0.,0.,0.

        latitude = float(location_line[6])
        longitude = float(location_line[7])
        time_zone = float(location_line[8])
        altitude = float(location_line[9])

        return (latitude, longitude, time_zone, altitude)

    except FileNotFoundError:
        print(f"Error: No se encontró el archivo en la ruta especificada: {epw_file_path}")
        return 0.,0.,0.,0.
    except (IndexError, ValueError) as e:
        print(f"Error: Formato de archivo EPW inválido. No se pudo analizar la información de ubicación. Detalles: {e}")
        return 0.,0.,0.,0.
    
def select_epjson_model(war_list: List[float]) -> Tuple[int, List[int]]:
    """
    Select the epjson model based on the window area ratios.

    Args:
        war_list (List[float]): List of window area ratios.

    Returns:
        int: Selected model number.
    
    Example:
        war_list = [0.2, 0.3, 0.1, 0.4]  # Example window area ratios for [n, e, s, w]
        model_number = select_epjson_model(war_list)
        print(f"Selected model number: {model_number}")
    """
    model_window_configs = {
        "1": [1,1,1,1], # all windows: [n,e,s,w]
        "2": [1,1,1,0],
        "3": [1,1,0,1],
        "4": [1,0,1,1],
        "5": [0,1,1,1],
        "6": [1,1,0,0],
        "7": [1,0,1,0],
        "8": [0,1,1,0],
        "9": [1,0,0,1],
        "10": [0,1,0,1],
        "11": [0,0,1,1],
        "12": [1,0,0,0],
        "13": [0,1,0,0],
        "14": [0,0,1,0],
        "15": [0,0,0,1],
    }
    
    war_list_bin = [1 if war > 0 else 0 for war in war_list]
    for model, config in model_window_configs.items():
        if config == war_list_bin:
            logger.info(f"EpisodeFunctionUtils: Selected model {model} for window area ratios {war_list}")
            return int(model), config
    
    logger.warning("EpisodeFunctionUtils: No matching model found for the given window area ratios. Defaulting to model 1.")
    return 1, [1,1,1,1]  # Default model if no match found

def get_random_parameter(min: float, max: float, step: float) -> float:
    """
    Generate a random parameter value within a specified range and step.

    Args:
        min (float): Minimum value of the range.
        max (float): Maximum value of the range.
        step (float): Step size for the values.

    Returns:
        float: A random value within the specified range and step.
    
    Example:
        random_value = get_random_parameter(0.0, 1.0, 0.1)
        print(f"Random value: {random_value}")
    """
    num_steps = int((max - min) / step) + 1
    random_step = np.random.randint(0, num_steps)
    return min + random_step * step


"""
Agents
=======

This module contains classes for representing and manipulating agents in the 
environment. The agents are responsible for taking actions in the environment 
following a specified policy that responses to the current state of the environment.

In this module, you will find:

- ``AgentSpec``: The main class for defining agents, including their observation, 
    filter, action, action_mapper, and reward specifications.
- ``ObservationSpec``: Defines the observation space for the agent.
- ``FilterSpec``: Defines filters to preprocess observations before they are fed to the agent.
- ``ActionSpec``: Defines the action space and actuators for the agent.
- ``ActionMapperSpec``: Defines ActionMappers that determine how the agent 
    should transform policy actions into actutators an actions.
- ``RewardSpec``: Defines the reward function for the agent.

.. note:: ``Filter`` must to be coordinated with ``Connector``.

Additionally, you will find base classes and some applications for ``Filters``, ``Rewards``, and ``ActionMappers``, 
which are essential parts of an agent in ``eprllib``.

"""

from typing import Dict, Any


SIMULATION_PARAMETERS: Dict[str, bool] = {
    'actual_date_time': False,
    'actual_time': False,
    'current_time': False,
    'day_of_month': False,
    'day_of_week': False,
    'day_of_year': False,
    'holiday_index': False,
    'hour': False,
    'minutes': False,
    'month': False,
    'num_time_steps_in_hour': False,
    'year': False,
    'is_raining': False,
    'sun_is_up': False,
    'today_weather_albedo_at_time': False,
    'today_weather_beam_solar_at_time': False,
    'today_weather_diffuse_solar_at_time': False,
    'today_weather_horizontal_ir_at_time': False,
    'today_weather_is_raining_at_time': False,
    'today_weather_is_snowing_at_time': False,
    'today_weather_liquid_precipitation_at_time': False,
    'today_weather_outdoor_barometric_pressure_at_time': False,
    'today_weather_outdoor_dew_point_at_time': False,
    'today_weather_outdoor_dry_bulb_at_time': False,
    'today_weather_outdoor_relative_humidity_at_time': False,
    'today_weather_sky_temperature_at_time': False,
    'today_weather_wind_direction_at_time': False,
    'today_weather_wind_speed_at_time': False,
    'tomorrow_weather_albedo_at_time': False,
    'tomorrow_weather_beam_solar_at_time': False,
    'tomorrow_weather_diffuse_solar_at_time': False,
    'tomorrow_weather_horizontal_ir_at_time': False,
    'tomorrow_weather_is_raining_at_time': False,
    'tomorrow_weather_is_snowing_at_time': False,
    'tomorrow_weather_liquid_precipitation_at_time': False,
    'tomorrow_weather_outdoor_barometric_pressure_at_time': False,
    'tomorrow_weather_outdoor_dew_point_at_time': False,
    'tomorrow_weather_outdoor_dry_bulb_at_time': False,
    'tomorrow_weather_outdoor_relative_humidity_at_time': False,
    'tomorrow_weather_sky_temperature_at_time': False,
    'tomorrow_weather_wind_direction_at_time': False,
    'tomorrow_weather_wind_speed_at_time': False,
}

ZONE_SIMULATION_PARAMETERS: Dict[str, bool] = {
    'system_time_step': False,
    'zone_time_step': False,
    'zone_time_step_number': False,
}

PREDICTION_VARIABLES: Dict[str, bool] = {
    'albedo': False,
    'beam_solar': False,
    'diffuse_solar': False,
    'horizontal_ir': False,
    'is_raining': False,
    'is_snowing': False,
    'liquid_precipitation': False,
    'outdoor_barometric_pressure': False,
    'outdoor_dew_point': False,
    'outdoor_dry_bulb': False,
    'outdoor_relative_humidity': False,
    'sky_temperature': False,
    'wind_direction': False,
    'wind_speed': False,
}

PREDICTION_HOURS: int = 24

# --- CONFIGURING USER PROFILES BY ZONE ---
# Each profile now distinguishes between day and night zones.
# The sum of occupants in both zones at a given hour reflects the distribution of people.

OCCUPATION_PROFILES: Dict[str, Dict[str, Any]] = {
    "Office schedule": {
        "total_people": 1,
        "zone_daytime": {
            "weekdays": [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0],
            "weekends": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        },
        "zone_nightly": {
            "weekdays": [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1],
            "weekends": [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1],
        },
    },
    "Single with an office job": {
        "total_people": 1,
        "zone_daytime": {
            "weekdays": [0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0],
            "weekends": [0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
        },
        "zone_nightly": {
            "weekdays": [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1],
            "weekends": [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1],
        },
    },
    "Typical family, office job": { # 2 adults, 2 children
        "total_people": 1,
        "zone_daytime": {
            "weekdays": [0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
            "weekends": [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
        },
        "zone_nightly": {
            "weekdays": [1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
            "weekends": [1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1],
        },
    },
    "Always present": {
        "total_people": 1,
        "zone_daytime": {
            "weekdays": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            "weekends": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        },
        "zone_nightly": {
            "weekdays": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            "weekends": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
        },
    },
    # More profiles can be added here with the same structure.
}

# Validation list for allowed user types.
VALID_USER_TYPES = list(OCCUPATION_PROFILES.keys())
VALID_ZONE_TYPES = ["daytime", "nightly"]



"""
Default Episode
================

This module contains the default implementation of the episode functions for the EnergyPlus environment.
"""
from eprllib.Episodes.BaseEpisode import BaseEpisode
from eprllib.Utils.annotations import override

class DefaultEpisode(BaseEpisode):
    """
    This class provides the default implementation of the episode functions for the EnergyPlus environment.
    It inherits from the BaseEpisode class.
    """
    @override(BaseEpisode)
    def setup(self) -> None:
        pass



"""
Defining agents
================

This module implements the classes to define agents. Agents are defined by the ``AgentSpec`` class. This class
contains the observation, filter, action, action_mapper, and reward specifications. The observation is defined by the
``ObservationSpec`` class. The filter is defined by the ``FilterSpec`` class. The action is defined by the ``ActionSpec`` class.
The action_mapper is defined by the ``ActionMapperSpec`` class. The reward is defined by the ``RewardSpec`` class.

The ``AgentSpec`` class has a method called ``build`` that is used to build the ``AgentSpec`` object. This method is used to
validate the properties of the object and to return the object as a dictionary. It is used internally when you build
the environment to provide it to RLlib.
"""
from typing import Dict, Any, Optional

from eprllib.Agents.Rewards.RewardSpec import RewardSpec
from eprllib.Agents.Filters.FilterSpec import FilterSpec
from eprllib.Agents.ActionSpec import ActionSpec
from eprllib.Agents.ActionMappers.ActionMapperSpec import ActionMapperSpec
from eprllib.Agents.ObservationSpec import ObservationSpec
from eprllib import logger

class AgentSpec:
    """
    AgentSpec is the base class for an agent specification to safe configuration of the object.
    """
    observation: Optional[ObservationSpec|Dict[str, Any]] = None
    filter: Optional[FilterSpec|Dict[str, Any]] = None
    action: Optional[ActionSpec|Dict[str, Any]] = None
    action_mapper: Optional[ActionMapperSpec|Dict[str, Any]] = None
    reward: Optional[RewardSpec|Dict[str, Any]] = None
    
    def __init__(
        self,
        observation: Optional[ObservationSpec|Dict[str, Any]] = None,
        filter: Optional[FilterSpec|Dict[str, Any]] = None,
        action: Optional[ActionSpec|Dict[str, Any]] = None,
        action_mapper: Optional[ActionMapperSpec|Dict[str, Any]] = None,
        reward: Optional[RewardSpec|Dict[str, Any]] = None,
        **kwargs: Any
        ) -> None:
        """
        Contruction method for the AgentSpec class.

        Args:
            observation (ObservationSpec, optional): Defines the observation of the agent using
            the ObservationSpec class or a Dict. Defaults to NotImplemented.
            filter (FilterSpec, optional): Defines the filter for this agent using FilterSpec or a Dict. Defaults to None.
            action (ActionSpec, optional): Defines the action characteristics of the agent using ActionSpec or a Dict. Defaults to NotImplemented.
            action_mapper (ActionMapperSpec, optional): Defines the action_mapper for the agent using ActionMapperSpec or a Dict. Defaults to None.
            reward (RewardSpec, optional): Defines the reward elements of the agent using RewardSpec or a Dict. Defaults to NotImplemented.

        Raises:
            NotImplementedError: _description_
            NotImplementedError: _description_
            NotImplementedError: _description_
        """
        if observation is not None:
            self.observation = observation
        else:
            logger.info("AgentSpec: No observation defined. Using default observation.")
            self.observation = ObservationSpec()
        
        if filter is not None:
            self.filter = filter
        else:
            logger.info("AgentSpec: No filter defined. Using default filter.")
            self.filter = FilterSpec()
            
        if action is not None:
            self.action = action
        else:
            logger.info("AgentSpec: No action defined. Using default action.")
            self.action = ActionSpec()
            
        if action_mapper is not None:
            self.action_mapper = action_mapper
        else:
            logger.info("AgentSpec: No action_mapper defined. Using default action_mapper.")
            self.action_mapper = ActionMapperSpec()
            
        if reward is not None:
            self.reward = reward
        else:
            logger.info("AgentSpec: No reward defined. Using default reward.")
            self.reward = RewardSpec()
        
        for key, value in kwargs.items():
            setattr(self, key, value)
    
    def __getitem__(self, key:str):
        return getattr(self, key)

    def __setitem__(self, key:str, value:Any):
        setattr(self, key, value)
        
        
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the AgentSpec object.
        """
        if isinstance(self.observation, ObservationSpec):
            self.observation = self.observation.build()
        else:
            msg = f"AgentSpec: The observation must be defined as an ObservationSpec object but {type(self.observation)} was given."
            logger.error(msg)
            raise ValueError(msg)
        if isinstance(self.filter, FilterSpec):
            self.filter = self.filter.build()
        else:
            msg = f"AgentSpec: The filter must be defined as a FilterSpec object but {type(self.filter)} was given."
            logger.error(msg)
            raise ValueError(msg)
        if isinstance(self.action, ActionSpec):
            self.action = self.action.build()
        else:
            msg = f"AgentSpec: The action must be defined as an ActionSpec object but {type(self.action)} was given."
            logger.error(msg)
            raise ValueError(msg)
        if isinstance(self.action_mapper, ActionMapperSpec):
            self.action_mapper = self.action_mapper.build()
        else:
            msg = f"AgentSpec: The action_mapper must be defined as a ActionMapperSpec object but {type(self.action_mapper)} was given."
            logger.error(msg)
            raise ValueError(msg)
        if isinstance(self.reward, RewardSpec):
            self.reward = self.reward.build()
        else:
            msg = f"AgentSpec: The reward must be defined as a RewardSpec object but {type(self.reward)} was given."
            logger.error(msg)
            raise ValueError(msg)

        return vars(self)



"""
Add EnergyPlus API path to sys.path for eprllib
================================================

This function manages the EnergyPlus API path for eprllib. It attempts to auto-detect the 
EnergyPlus installation based on the operating system and a provided version string. If a 
valid path is found, it adds it to `sys.path` for use by eprllib.
"""

from typing import Optional
import sys
from eprllib import logger
from eprllib.version import ep_version_list


def EP_API_add_path(path: Optional[str] = None) -> str:
    """
    Manages EnergyPlus paths for eprllib.

    If a 'path' argument is provided, it's used directly as the EnergyPlus installation path
    after validation.
    Otherwise, this method auto-detects installed EnergyPlus versions based on `LIST_OF_VERSIONS`.
    - If multiple versions are found, the latest one (as per `LIST_OF_VERSIONS`) is chosen,
      and a message is printed.
    - If no version is found, an error message is printed, and the program exits.

    The selected EnergyPlus installation (either user-provided or auto-detected) is then
    copied to a unique temporary directory. This temporary copy's path is added to `sys.path`,
    allowing isolated EnergyPlus environments for parallel execution.
    Temporary directories are registered for cleanup on program exit using `atexit`.

    Args:
        path (Optional[str], optional): Full path to an EnergyPlus installation directory.
            If None, auto-detection is performed. Defaults to None.

    Returns:
        str: The path to the temporary EnergyPlus copy that was added to `sys.path`.

    Raises:
        FileNotFoundError: If 'path' is provided but does not exist or is not a directory.
        RuntimeError: If auto-detection fails to find any suitable EnergyPlus installation,
                      or if copying the installation to a temporary directory fails.
    """
    logger.debug("EnvConfigUtils: Attempting to auto-detect EnergyPlus installation...")
    os_platform = sys.platform
    original_ep_path: Optional[str] = None
    
    # Check that the provided path is valid.
    if path not in ep_version_list:
        msg = f"EnvConfigUtils: Invalid EnergyPlus version provided: {path}. " \
              f"Valid versions are: {ep_version_list}"
        logger.error(msg)
        raise ValueError(msg)
    
    if os_platform.startswith("linux"):  # Covers "linux" and "linux2"
        original_ep_path = f"/usr/local/EnergyPlus-{path}"
    elif os_platform == "win32":
        original_ep_path = f"C:/EnergyPlusV{path}"
    elif os_platform == "darwin":
        original_ep_path = f"/Applications/EnergyPlus-{path}"

    if original_ep_path is not None:
        if original_ep_path not in sys.path:
            sys.path.insert(0, original_ep_path)
            logger.debug(f"EnvConfigUtils: EnergyPlus API path added to sys.path: {original_ep_path}")
        else:
            logger.debug(f"EnvConfigUtils: EnergyPlus API path already in sys.path: {original_ep_path}")
        
        return original_ep_path
    
    else:
        logger.error(f"EnvConfigUtils: Warning: EnergyPlus auto-detection is not configured for this OS: {os_platform}. "
                "Please provide the path manually if detection fails.")
        raise RuntimeError(f"EnergyPlus auto-detection failed for OS: {os_platform}. "
                           "Please provide the path manually or ensure EnergyPlus is installed correctly.")
        



"""
Setpoint ActionMappers
========================

This module contains classes to implement setpoint ActionMappers for controlling actuators in the environment.
"""
import gymnasium as gym
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override


class DualSetpointDiscreteAndAvailabilityActionMapper(BaseActionMapper):
    
    @override(BaseActionMapper)
    def setup(self):
        # Here you have access to the self.action_mapper_config and self.agent_name
        
        # The name of the actuator that will be controlled by the agent. 
        # This will be obtained from the actuators_config in the environment configuration file.
        self.actuator_for_cooling: Optional[str] = None
        self.actuator_for_heating: Optional[str] = None
        self.availability_actuator: Optional[str] = None

        # Here we use the config dict to provide the action space dimension.
        self.action_space_dim: int = self.action_mapper_config.get("action_space_dim", 11)
        if self.action_space_dim < 3:
            raise ValueError("The action_space_dim must be greater than or equal to 3.")
        
        temperature_range: Tuple[float|int, float|int] = self.action_mapper_config['temperature_range']
        self.temperature_range_max: float|int = max(temperature_range)
        self.temperature_range_min: float|int = min(temperature_range)
        self.deadband: int = self.action_mapper_config.get('deadband', 2)
        
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(self.action_space_dim)
    
    
    @override(BaseActionMapper)
    def actuator_names(
        self, 
        actuators_config: Dict[str, Tuple[str,str,str]]
        ) -> None:
        """
        This method is used to transform the agent dict action to actuator dict action. Consider that
        one agent could manage more than one actuator. For that reason, it is important to transform the
        action dict to actuator dict actions.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        
        # Here the name of the actuator is obtained from the actuators_config 
        # in the environment configuration file.
        self.actuator_for_cooling = get_actuator_name(
            self.agent_name,
            actuators_config["Cooling Actuator"][0],
            actuators_config["Cooling Actuator"][1],
            actuators_config["Cooling Actuator"][2]
        )
        self.actuator_for_heating = get_actuator_name(
            self.agent_name,
            actuators_config["Heating Actuator"][0],
            actuators_config["Heating Actuator"][1],
            actuators_config["Heating Actuator"][2]
        )
        self.availability_actuator = get_actuator_name(
            self.agent_name,
            actuators_config["Availability Actuator"][0],
            actuators_config["Availability Actuator"][1],
            actuators_config["Availability Actuator"][2]
        )
    
    
    @override(BaseActionMapper)
    def _agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """ 
        assert self.actuator_for_cooling in actuators, "Cooling actuator name has not been initialized."
        assert self.actuator_for_heating in actuators, "Heating actuator name has not been initialized."
        assert self.availability_actuator in actuators, "Availability actuator name has not been initialized."
        
        if action == 0:
            return {
                self.actuator_for_heating: float(self.temperature_range_min),
                self.actuator_for_cooling: float(self.temperature_range_min + self.deadband),
                self.availability_actuator: 0
            }
        
        else:
            return {
                self.actuator_for_heating: float(self.temperature_range_min + ((action-1)/(self.action_space_dim-2))*(self.temperature_range_max - self.temperature_range_min - self.deadband)),
                self.actuator_for_cooling: float(self.temperature_range_min + ((action-1)/(self.action_space_dim-2))*(self.temperature_range_max - self.temperature_range_min - self.deadband) + self.deadband),
                self.availability_actuator: 1
            }




"""
Exhaust Fan ActionMappers
=============================

This module contains classes to implement ActionMappers for controlling exhaust fan actuators in the environment.
"""
import gymnasium as gym
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override
from eprllib import logger

class ExhaustFanActionMapper(BaseActionMapper):
    
    @override(BaseActionMapper)
    def setup(self):
        
        # The name of the actuator that will be controlled by the agent. 
        # This will be obtained from the actuators_config in the environment configuration file.
        self.exhaust_fan_actuator: Optional[str] = None
        
        # Here we use the config dict to provide the action space dimension.
        self.action_space_dim: int = self.action_mapper_config.get("action_space_dim", 11)
        
        # Here we use the config dict to provide the modes of the exhaust fan.
        self.modes: List[float] = self.action_mapper_config['modes']
        
        #  Check if the lenght of the modes are larger than 11 (that is the action space for this class).
        if len(self.modes) > 11:
            msg = f"The lenght of the modes must be less than 11. The lenght of the modes is {len(self.modes)}."
            logger.error(msg)
            raise ValueError(msg)
        
        # Check that all the elements in self.modes list are floats in the range [0, 1].
        for mode in self.modes:
            if mode < 0 or mode > 1:
                msg = f"The mode {mode} is not in the range [0, 1]."
                logger.error(msg)
                raise ValueError(msg)
    
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Discrete(self.action_space_dim)
    
    
    @override(BaseActionMapper)
    def actuator_names(
        self, 
        actuators_config: Dict[str, Tuple[str,str,str]]
        ) -> None:
        """
        This method is used to transform the agent dict action to actuator dict action. Consider that
        one agent could manage more than one actuator. For that reason, it is important to transform the
        action dict to actuator dict actions.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        
        # Here the name of the actuator is obtained from the actuators_config 
        # in the environment configuration file.
        self.exhaust_fan_actuator = get_actuator_name(
            self.agent_name,
            actuators_config["Exhaust Fan Actuator"][0],
            actuators_config["Exhaust Fan Actuator"][1],
            actuators_config["Exhaust Fan Actuator"][2]
        )
        
        
    @override(BaseActionMapper)
    def _agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
            
        Raises:
            ValueError: If the actuator is not in the list of actuators.
        """
        if self.exhaust_fan_actuator in actuators:
            return {self.exhaust_fan_actuator: self.modes[action]}
        else:
            return {}



"""
Air Mass Flow Rate ActionMappers
===================================

An actuator called “Ideal Loads Air System” is available with control types called “Air Mass Flow
Rate” (supply air), “Outdoor Air Mass Flow Rate,” “Air Temperature,” and “Air Humidity Ratio.”
These are available in models that use the ideal loads air system, formerly known as purchased
air. The units are kg/s for mass flow rate, C for temperature and kgWater/kgDryAir for humidity
ratio. The unique identifier is the user-defined name of the ZoneHVAC:IdealLoadsAirSystem input
object.
For Air Temperature and Air Humidity Ratio, the overrides are absolute. They are applied after
all other limits have been checked. For mass flow rate, the overrides are not absolute,the internal
controls will still apply the capacity and flow rate limits if defined in the input object. The EMS
override will be ignored if the ideal loads system is off (the availability schedule value is zero or it
has been forced “off” by an availability manager). If both the Air Mass Flow Rate and Outdoor Air
Mass Flow Rate are overridden, the Outdoor Air Mass Flow Rate will not be allowed to be greater
than the override value for Air Mass Flow Rate.
"""
import gymnasium as gym
import numpy as np
from typing import Any, Dict, List, Tuple, Optional
from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Utils.observation_utils import get_actuator_name
from eprllib.Utils.annotations import override

class AirMassFlowRateActionMapper(BaseActionMapper):
    
    @override(BaseActionMapper)
    def setup(self) -> None:
        """
        This class implements the “Air Mass Flow Rate” (supply air) action function.

        Args:
            action_mapper_config (Dict[str, Any]): The configuration of the action function.
            It should contain the following keys:
                - air_mass_flow_rate_range (Tuple[int, int]): The range for the air mass flow rate.
                - ideal_loads_air_system_actuator (Tuple[str, str, str]): The configuration for the actuator.
                
        """
        # The name of the actuator that will be controlled by the agent. 
        # This will be obtained from the actuators_config in the environment configuration file.
        self.air_mass_flow_rate_actuator: Optional[str] = None
        
        # Here we use the config dict to provide the range of the air mass flow rate.
        air_mass_flow_rate_range: Tuple[float|int, float|int] = self.action_mapper_config['air_mass_flow_rate_range']
        self.air_mass_flow_rate_range_max: float|int = max(air_mass_flow_rate_range)
        self.air_mass_flow_rate_range_min: float|int = min(air_mass_flow_rate_range)
        
    
    @override(BaseActionMapper)    
    def get_action_space_dim(self) -> gym.Space[Any]:
        """
        Get the action space of the environment.

        Returns:
            gym.Space: Action space of the environment.
        """
        return gym.spaces.Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)
    
    
    @override(BaseActionMapper)
    def actuator_names(
        self, 
        actuators_config: Dict[str, Tuple[str,str,str]]
        ) -> None:
        """
        This method is used to transform the agent dict action to actuator dict action. Consider that
        one agent could manage more than one actuator. For that reason, it is important to transform the
        action dict to actuator dict actions.

        Args:
            action (Any): The action to be transformed.
            actuators (List[str]): List of actuators controlled by the agent.

        Returns:
            Dict[str, Any]: Transformed actions for the actuators.
        """
        
        self.air_mass_flow_rate_actuator = get_actuator_name(
            self.agent_name,
            actuators_config['Air Mass Flow Rate Actuator'][0],
            actuators_config['Air Mass Flow Rate Actuator'][1],
            actuators_config['Air Mass Flow Rate Actuator'][2]
        )
    
    
    @override(BaseActionMapper)
    def _agent_to_actuator_action(self, action: Any, actuators: List[str]) -> Dict[str,Any]:
        """
        This method is used to transform the agent action to actuator dict action. Consider that
        one agent may manage more than one actuator.

        Args:
            action (Any): The agent action, normally an int of float.
            actuators (List[str]): The list of actuator names that the agent manage.

        Return:
            Dict[str, Any]: Dictionary with the actions for the actuators.
        """
        if self.air_mass_flow_rate_actuator not in actuators:
            return {}
        
        else:
            if action <= 0.0:
                return {self.air_mass_flow_rate_actuator: float(self.air_mass_flow_rate_range_min)}
            
            elif action >= 1.0:
                return {self.air_mass_flow_rate_actuator: float(self.air_mass_flow_rate_range_max)}
            
            else:
                return {self.air_mass_flow_rate_actuator: float(self.air_mass_flow_rate_range_min + (action)*(self.air_mass_flow_rate_range_max - self.air_mass_flow_rate_range_min))}



"""
Runner Base Class
==================

This script contain the EnergyPlus Runner that execute EnergyPlus from its 
Python API in the version 24.2.0.
"""
import os
import sys
import threading
import time
import numpy as np
from queue import Queue
from ctypes import c_void_p
from typing import Any, Dict, List, Optional, Tuple, cast
from types import FunctionType

from eprllib.Agents.ActionMappers.BaseActionMapper import BaseActionMapper
from eprllib.Agents.Filters.BaseFilter import BaseFilter
from eprllib.Connectors.BaseConnector import BaseConnector
from eprllib.Utils.observation_utils import (
    get_actuator_name,
    get_internal_variable_name,
    get_meter_name,
    get_parameter_name,
    get_other_obs_name,
    get_variable_name,
    get_parameter_prediction_name,
    get_user_occupation_forecast_name
)
from eprllib.Utils.env_utils import calculate_occupancy, calculate_occupancy_forecast
from eprllib.Utils.add_ep_to_path import EP_API_add_path
from eprllib import logger, EP_VERSION


class EnvironmentRunner:
    """
    This object have the particularity of `start` EnergyPlus, `_collect_obs` and `_send_actions` to
    send it trhougt queue to the EnergyPlus Environment thread.
    """
    observation_config: Dict[str, Any]
    generals_config: Dict[str, Any]
    actuator_config: Dict[str, Any]
    episode: int
    obs_queue: Queue[Any]
    act_queue: Queue[Any]
    infos_queue: Queue[Any]
    agents: List[str] = []
    obs_event: threading.Event= threading.Event()
    act_event: threading.Event= threading.Event()
    infos_event: threading.Event= threading.Event()
    energyplus_exec_thread: Optional[threading.Thread] = None
    energyplus_state: Optional[c_void_p] = None
    sim_results: int = 0
    initialized: bool = False
    init_handles: bool = False
    simulation_complete: bool = False
    first_observation: bool = True
    obs: Dict[str, Any] = {}
    infos: Dict[str, Dict[str, Any]]
    unique_id: float = time.time()
    is_last_timestep: bool = False
    occupancy_next_timestep: float = 0.
    action_mapper: Dict[str, BaseActionMapper]
    filter_fn: Dict[str, BaseFilter]
    connector_fn: BaseConnector
    agent_variables_and_handles: Dict[str, Any] = {}
    internal_actuators: List[Tuple[str,str,str]] = []
    internal_variables_and_handles: Dict[str, Any] = {}
    energyplus_exec_thread_timeout: int = 1
    ep_worker_path: Optional[str]
    ep_version: str = EP_VERSION
    
    def __init__(
        self,
        episode: int,
        obs_queue: Queue[Any],
        act_queue: Queue[Any],
        infos_queue: Queue[Any],
        agents: List[str],
        filter_fn: Dict[str,BaseFilter],
        action_mapper: Dict[str, BaseActionMapper],
        connector_fn: BaseConnector,
        observation_config: Dict[str, Any],
        generals_config: Dict[str, Any],
        actuator_config: Dict[str, Any],
        ep_worker_path: Optional[str],
        ep_version: str = "24-2-0"
        ) -> None:
        """
        Initializes the BaseRunner object.

        Args:
            env_config (Dict[str, Any]): Configuration settings for the environment.
            episode (int): The current episode number.
            obs_queue (Queue): Queue for sending observations.
            act_queue (Queue): Queue for receiving actions.
            infos_queue (Queue): Queue for sending additional information.
            agents (List): List of agents in the environment.
            filter_fn (Dict[str, BaseFilter]): Dictionary of filter functions for each agent.
            action_mapper (Dict[str, BaseActionMapper]): Dictionary of action_mapper functions for each agent.
            connector_fn (BaseConnector): Connector function for combining agent observations.
            observation_config (Dict[str, Any]): Configuration for observations.
            generals_config (Dict[str, Any]): Configuration for general settings.
            actuator_config (Dict[str, Any]): Configuration for actuators.
        """
        # Asignation of variables.
        self.episode = episode
        self.obs_queue = obs_queue
        self.act_queue = act_queue
        self.infos_queue = infos_queue
        self.agents = agents
        self.observation_config = observation_config
        self.generals_config = generals_config
        self.actuator_config = actuator_config
        self.ep_worker_path = ep_worker_path

        if self.ep_worker_path is None:
            msg = f"EnvironmentRunner: WARNING: self.ep_worker_path is None. Try to have the path..."
            logger.warning(msg)
            # EnergyPlus Python API path adding
            try:
                self.ep_worker_path = EP_API_add_path(self.ep_version)
                logger.warning(f"EnvironmentRunner: The self.ep_worker_path was created as {self.ep_worker_path}.")
            except RuntimeError as e:
                logger.error(f"EnvironmentRunner: Failed to add EnergyPlus API path: {e}")
                exit(1)
        
        if self.ep_worker_path and self.ep_worker_path not in sys.path:
            sys.path.insert(0, self.ep_worker_path)
        
        from pyenergyplus.api import EnergyPlusAPI
        self.api = EnergyPlusAPI()
        
        logger.debug(f"EnvironmentRunner: The EnergyPlus application was imported successfully from {self.ep_worker_path}.")
        
        # The queue events are generated (To sure the coordination with EnergyPlusEnvironment).
        self.obs_event = threading.Event()
        self.act_event = threading.Event()
        self.infos_event = threading.Event()
        
        # Variables to be used in this thread.
        self.energyplus_exec_thread: Optional[threading.Thread] = None
        self.energyplus_state: Optional[c_void_p] = None
        self.sim_results: int = 0
        self.initialized = False
        self.init_handles = False
        self.simulation_complete = False
        self.first_observation = True
        self.obs = {}
        self.infos = {agent: {} for agent in self.agents}
        self.unique_id = time.time()
        self.is_last_timestep: bool = False
        self.occupancy_next_timestep: float = 0.
        self.num_time_steps_in_hour: int = 0
        self.energyplus_exec_thread_timeout: int = 1
        
        # create a variable to save the obs dict key to use in posprocess
        # self.obs_keys = []
        # self.infos_keys = []
        
        # Define the action and observation functions.
        self.action_mapper = action_mapper
        self.filter_fn = filter_fn
        self.connector_fn = connector_fn
        
        # Declaration of variables, meters and actuators to use in the simulation. Handles
        # are used in _init_handle method.
        self.agent_variables_and_handles: Dict[str, Any] = {}
        for agent in self.agents:
            variables, variables_handles = self.set_variables(agent)
            internal_variables, internal_variables_handles = self.set_internal_variables(agent)
            meters, meters_handles = self.set_meters(agent)
            actuators, actuators_handles = self.set_actuators(agent)
            self.agent_variables_and_handles.update({
                f"{agent}_variables": [variables, variables_handles],
                f"{agent}_internal_variables": [internal_variables, internal_variables_handles],
                f"{agent}_meters": [meters, meters_handles],
                f"{agent}_actuators": [actuators, actuators_handles]
            })
        
        # Declaration of internal actuators for funtions and not agents.
        internal_actuators: List[Tuple[str,str,str]] = []
        for agent in self.agents:
            # Occupation funtion.
            if self.observation_config[agent]["user_occupation_function"]:
                occupation_schedule = self.observation_config[agent].get("occupation_schedule", None)
                if occupation_schedule is not None:
                    internal_actuators.append(tuple(occupation_schedule))
                else:
                    msg = f"EnvironmentRunner: The schedule must be defined in the observation config for agent {agent}."
                    logger.error(msg)
                    raise ValueError(msg)
            
        # delete duplicate actuators in actuators_list
        internal_actuators = list(set(internal_actuators))
        
        self.internal_variables_and_handles: Dict[str, Any] = {}
        int_actuators, int_actuators_handles = self.set_internal_actuators(internal_actuators)
        self.internal_variables_and_handles.update({
                "internal_actuators": [int_actuators, int_actuators_handles]
            })
    
    def progress_handler(self, progress: int) -> None:
        # print(f"Simulation progress: {progress}%")
        if progress >= 99:
            self.is_last_timestep = True
            
    def start(self) -> None:
        """
        Starts the EnergyPlus simulation.
        """
        # Start a new EnergyPlus state (condition for execute EnergyPlus Python API).
        try:
            self.energyplus_state = self.api.state_manager.new_state()
            assert self.energyplus_state is not None, "EnvironmentRunner: Failed to create EnergyPlus state"
        except Exception as e:
            msg = f"EnvironmentRunner: Error creating EnergyPlus state: {e}"
            logger.error(f"msg")
            raise RuntimeError(msg)
        
        # Request variables.
        for agent in self.agents:
            if self.observation_config[agent]["variables"] is not None:
                for variable in self.observation_config[agent]["variables"]:
                    self.api.exchange.request_variable(
                        self.energyplus_state,
                        variable_name = variable[0],
                        variable_key = variable[1]
                    )
                    # print(f"Variable {get_variable_name(agent,variable[0],variable[1])} requested.")
        self.api.runtime.callback_begin_zone_timestep_after_init_heat_balance(self.energyplus_state, cast(FunctionType, self._send_actions))
        self.api.runtime.callback_end_zone_timestep_after_zone_reporting(self.energyplus_state, cast(FunctionType, self._collect_obs))
        self.api.runtime.callback_progress(self.energyplus_state, cast(FunctionType, self.progress_handler))
        self.api.runtime.set_console_output_status(self.energyplus_state, self.generals_config['ep_terminal_output']) # type: ignore
        
        def _run_energyplus():
            """
            Run EnergyPlus in a non-blocking way with Threads.
            """
            try:
                cmd_args = self.make_eplus_args()
                logger.info(f"EnvironmentRunner: running EnergyPlus with args: {cmd_args}")
                assert self.energyplus_state is not None, "EnvironmentRunner: EnergyPlus state is None."
                self.sim_results = self.api.runtime.run_energyplus(self.energyplus_state, cmd_args)
            except Exception as e:
                msg = f"EnvironmentRunner: Error running EnergyPlus: {e}"
                logger.error(msg)
                self.sim_results = -1
            finally:
                self.simulation_complete = True
            
        self.energyplus_exec_thread = threading.Thread(
            target=_run_energyplus,
            args=()
        )
        # Here the thread is divide in two.
        self.energyplus_exec_thread.start()

    def _collect_obs(
        self,
        state_argument: c_void_p
        ) -> None:
        """
        EnergyPlus callback that collects output variables, meters and actuator actions
        values and enqueue them to the EnergyPlus Environment thread.
        """
        # To not perform observations when the callbacks and the 
        # warming period are not complete.
        if not self._init_callback(state_argument) or self.simulation_complete:
            return
        
        dict_agents_obs: Dict[str, Any] = {agent: {} for agent in self.agents}
        self.infos: Dict[str, Dict[str, Any]] = {agent: {} for agent in self.agents}
        
        # Agents observe: site state, thermal zone state (only the one that it belong), specific object variables 
        # and meters, and others parameters assigned as True in the env_config.observation object.
        # Get the state of the actuators.
        agent_states: Dict[str, Dict[str, Any]] = {agent: {} for agent in self.agents}
        for agent in self.agents:
            agent_states[agent].update(self.get_variables_state(state_argument, agent))
            agent_states[agent].update(self.get_internal_variables_state(state_argument, agent))
            agent_states[agent].update(self.get_meters_state(state_argument, agent))
            agent_states[agent].update(self.get_simulation_parameters_values(state_argument, agent))
            agent_states[agent].update(self.get_zone_simulation_parameters_values(state_argument, agent))
            agent_states[agent].update(self.get_weather_prediction(state_argument, agent))
            agent_states[agent].update(self.get_actuators_state(state_argument, agent))
            agent_states[agent].update(self.get_other_obs(agent))
            agent_states[agent].update(self.get_user_occupation_forecast(state_argument, agent))
        
        # Filter the observation.
        for agent in self.agents:
            dict_agents_obs.update({
                agent: self.filter_fn[agent].get_filtered_obs(
                    agent_states[agent]
                )})
        
        # First is send the observation of the top-level agents. If there is only one shape of agents, the _collect_obs
        # method is ended here. If there are more than one shape of agents, the action (goal) will be requested and the
        # observations of the next shape will be request. This is implemented until reach the lowest level of the hierarchy
        # if any.
        top_level_agents_obs, top_level_agents_infos, is_lowest_level = self.connector_fn.set_top_level_obs(
            agent_states,
            dict_agents_obs,
            self.infos,
            self.is_last_timestep
        )
        
        # Set the agents observation and infos to communicate with the EPEnv.
        self.obs_queue.put(top_level_agents_obs)
        self.obs_event.set()
        self.infos_queue.put(top_level_agents_infos)
        self.infos_event.set()
        
        # Implementation for hierarchical agents.
        if not is_lowest_level:
            while not is_lowest_level:
                # Wait for a goal selection
                event_flag = self.act_event.wait(self.generals_config["timeout"])
                if not event_flag:
                    # print(f"Timeout waiting for action from agent {agent}.")
                    return
                # Get the action from the EnergyPlusEnvironment `step` method.
                actions: Dict[str, int | float] = self.act_queue.get()
                goals: Dict[str, Optional[int|float]] = {agent: None for agent in actions.keys()}
                for agent in actions.keys():
                    goals.update({agent: self.action_mapper[agent].action_to_goal(actions[agent])})
                
                low_level_agents_obs, low_level_agents_infos, is_lowest_level = self.connector_fn.set_low_level_obs(
                    agent_states,
                    dict_agents_obs,
                    self.infos,
                    goals
                )
            
                # Set the agents observation and infos to communicate with the EPEnv.
                self.obs_queue.put(low_level_agents_obs)
                self.obs_event.set()
                self.infos_queue.put(low_level_agents_infos)
                self.infos_event.set()
        

    def _collect_first_obs(self, state_argument: c_void_p):
        """
        This method is used to collect only the first observation of the environment when the episode beggins.

        Args:
            state_argument (): EnergyPlus state pointer. This is created with `self.api.state_manager.new_state()`.
        """
        if self.first_observation:
            # print("Collecting the first observation.")
            self._collect_obs(state_argument)
            self.first_observation = False
        else:
            return
    
    def _send_actions(self, state_argument: c_void_p) -> None:
        """EnergyPlus callback that sets actuator value from last decided action
        """
          
        # To not perform actions if the callbacks and the warming period are not complete.
        if not self._init_callback(state_argument) or self.simulation_complete:
            return
        
        # If is the first timestep, obtain the first observation before to consult for an action
        if self.first_observation:
            self.num_time_steps_in_hour = self.api.exchange.num_time_steps_in_hour(state_argument)
            self._collect_first_obs(state_argument)
            
        # Wait for an action.
        event_flag = self.act_event.wait(self.generals_config["timeout"])
        if not event_flag:
            # print("Timeout waiting for action from agent.")
            return
        # Get the action from the EnergyPlusEnvironment `step` method.
        dict_action: Dict[str, Any] = self.act_queue.get()
        
        for agent in dict_action.keys():
            # Transform action must to consider the agents and actuators and transform agents actions to actuators actions,
            # considering that one agent could manage more than one actuator.
            actuator_list = [actuator for actuator in self.agent_variables_and_handles[f"{agent}_actuators"][0].keys()]
            
            actuator_actions = self.action_mapper[agent].agent_to_actuator_action(dict_action[agent], actuator_list)
            
            # Check if there is an actuator_dict_actions value equal to None.
            for actuator in actuator_list:
                if actuator_actions[actuator] is None:
                    msg = f"EnvironmentRunner: The actuator {actuator} has no action defined in the action dict for agent {agent}."
                    logger.error(msg)
                    raise ValueError(msg)
                
            # Perform the actions in EnergyPlus simulation.
            for actuator in actuator_list:
                action = self.action_mapper[agent].get_actuator_action(actuator_actions[actuator], actuator)
                self.api.exchange.set_actuator_value(
                    state=state_argument,
                    actuator_handle=self.agent_variables_and_handles[f"{agent}_actuators"][1][actuator],
                    actuator_value=action
                )
                
        for actuator in self.internal_variables_and_handles["internal_actuators"][0].keys():
            self.api.exchange.set_actuator_value(
                state=state_argument,
                actuator_handle=self.internal_variables_and_handles["internal_actuators"][1][actuator],
                actuator_value=self.occupancy_next_timestep
            )
            
    def set_variables(self, agent:str) -> Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]:
        """
        The EnergyPlus variables are defined in the environment configuration.

        Args:
            agent (str): The AgentID. 

        Returns:
            Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]: The environment variables and their handles.
        """
        # Define the emptly Dict to include variables names and handles. The handles Dict return as emptly Dict to be used latter.
        var_handles: Dict[str, int] = {}
        variables: Dict[str, Tuple [str, str]] = {}
        if self.observation_config[agent]["variables"] is not None:
            variables.update({
                get_variable_name(agent,variable[0],variable[1]): variable
                for variable 
                in self.observation_config[agent]["variables"]
            })
        return variables, var_handles

    def set_internal_variables(self, agent:str) -> Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]:
        """
        Set the internal variables to handle and get the values in the environment.

        Args:
            agent (str): The AgentID.

        Returns:
            Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]: The environment variables and their handles.
        """
        # Define the emptly Dict to include variables names and handles. The handles Dict return as emptly Dict to be used latter.
        var_handles: Dict[str, int] = {}
        variables: Dict[str, Tuple [str, str]] = {}
        if self.observation_config[agent]["internal_variables"] is not None:
            variables.update({
                get_internal_variable_name(agent,variable[0],variable[1]): variable
                for variable 
                in self.observation_config[agent]["internal_variables"]
            })
        return variables, var_handles

    def set_meters(self, agent:str) -> Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]:
        """The EnergyPlus meters are defined in the environment configuration.
        
        Args:
            agent (str): The AgentID.

        Returns:
            Tuple[Dict[str, Tuple [str, str]], Dict[str, int]]: The meters and their handles.
        """
        var_handles: Dict[str, int] = {}
        variables: Dict[str, Tuple [str, str]] = {}
        if self.observation_config[agent]["meters"] is not None:
            variables.update({
                get_meter_name(agent,variable): variable 
                for variable 
                in self.observation_config[agent]["meters"]
            })
        return variables, var_handles

    def set_actuators(self, agent:str) -> Tuple[Dict[str,Tuple[str,str,str]], Dict[str, int]]:
        """The EnergyPlus actuators are defined in the environment configuration.

        Args:
            agent (str): The AgentID.

        Returns:
            Tuple[Dict[str,Tuple[str,str,str]], Dict[str,int]]: The actuators and their handles.
        """
        actuators: Dict[str,Tuple[str,str,str]] = {}
        actuator_handles: Dict[str, int] = {}
        
        for actuator_config in self.actuator_config[agent]["actuators"].values():
            actuators.update({
                get_actuator_name(agent,actuator_config[0],actuator_config[1],actuator_config[2]): actuator_config
            })
        
        return actuators, actuator_handles
    
    def set_internal_actuators(self, internal_actuators: List[Tuple[str,str,str]]) -> Tuple[Dict[str,Tuple[str,str,str]], Dict[str, int]]:
        """The EnergyPlus actuators are defined in the environment configuration.

        Args:
            internal_actuators (List[Tuple[str,str,str]]): List of actuator to be commanded internally.

        Returns:
            Tuple[Dict[str,Tuple[str,str,str]], Dict[str,int]]: The actuators and their handles.
        """
        actuators: Dict[str,Tuple[str,str,str]] = {}
        actuator_handles: Dict[str, int] = {}
        
        for actuator_config in internal_actuators:
            actuators.update({
                get_actuator_name("internal",actuator_config[0],actuator_config[1],actuator_config[2]): actuator_config
            })
        
        return actuators, actuator_handles
    
    def get_variables_state(
        self,
        state_argument: c_void_p,
        agent: Optional[str] = None
        ) -> Dict[str,Any]:
        """This funtion takes the state_argument and return a dictionary with the agent variables values.

        Args:
            state_argument (): State argument from EnergyPlus callback.

        Returns:
            Dict[str,Any]: Agent variables values for the actual timestep.
        """
        if agent is None:
            msg = "EnvironmentRunner: The agent must be defined."
            logger.error(msg)
            raise ValueError(msg)
        variables: Dict[str,Any] = {
            key: self.api.exchange.get_variable_value(state_argument, handle)
            for key, handle
            in self.agent_variables_and_handles[f"{agent}_variables"][1].items()
        }        
        return variables
    
    def get_internal_variables_state(
        self,
        state_argument: c_void_p,
        agent: Optional[str] = None
        ) -> Dict[str,Any]:
        """Get the static variable values defined in the static variable dict names and handles.

        Args:
            state_argument (): EnergyPlus state.

        Returns:
            Dict[str,Any]: Static value dict values for the actual timestep.
        """
        if agent is None:
            msg = "EnvironmentRunner: The agent must be defined."
            logger.error(msg)
            raise ValueError(msg)
        variables: Dict[str,Any] = {
            key: self.api.exchange.get_internal_variable_value(state_argument, handle)
            for key, handle
            in self.agent_variables_and_handles[f"{agent}_internal_variables"][1].items()
        }
        return variables

    def get_meters_state(
        self,
        state_argument: c_void_p,
        agent: Optional[str] = None
        ) -> Dict[str,Any]:
        """Get the static variable values defined in the static variable dict names and handles.

        Args:
            state_argument (): EnergyPlus state.

        Returns:
            Dict[str,Any]: Static value dict values for the actual timestep.
        """
        if agent is None:
            msg = "EnvironmentRunner: The agent must be defined."
            logger.error(msg)
            raise ValueError(msg)
        
        variables: Dict[str,Any] = {
            key: self.api.exchange.get_meter_value(state_argument, handle)
            for key, handle
            in self.agent_variables_and_handles[f"{agent}_meters"][1].items()
        }
        return variables
        
    def get_actuators_state(
        self,
        state_argument: c_void_p,
        agent: Optional[str] = None
        ) -> Dict[str,Any]:
        """This funtion takes the state_argument and return a dictionary with the agent actuator values.

        Args:
            state_argument (): State argument from EnergyPlus callback.

        Returns:
            Dict[str,Any]: Agent actuator values for the actual timestep.
        """
        if agent is None:
            msg = "EnvironmentRunner: The agent must be defined."
            logger.error(msg)
            raise ValueError(msg)
        
        variables: Dict[str,Any] = {
            key: self.api.exchange.get_actuator_value(state_argument, handle)
            for key, handle
            in self.agent_variables_and_handles[f"{agent}_actuators"][1].items()
        }
        return variables
    
    def get_simulation_parameters_values(
        self,
        state_argument: c_void_p,
        agent: str
        ) -> Dict[str,Any]:
        """Get the simulation parameters values defined in the simulation parameter list.

        Args:
            state_argument (): EnergyPlus state.
            simulation_parameter_list (List): List of the simulation parameters implemented in the observation space.

        Returns:
            Dict[str,int|float]: Dict of parameter names as keys and parameter values as values.
        """
        # Get timestep variables that are needed as input for some data_exchange methods.
        hour = self.api.exchange.hour(state_argument)
        zone_time_step_number = self.api.exchange.zone_time_step_number(state_argument)
        
        # Dict with the variables names and methods as values.
        parameter_methods: Dict[str,Any] = {
            'actual_date_time': self.api.exchange.actual_date_time(state_argument), # Gets a simple sum of the values of the date/time function. Could be used in random seeding.
            'actual_time': self.api.exchange.actual_time(state_argument), # Gets a simple sum of the values of the time part of the date/time function. Could be used in random seeding.
            'current_time': self.api.exchange.current_time(state_argument), # Get the current time of day in hours, where current time represents the end time of the current time step.
            'day_of_month': self.api.exchange.day_of_month(state_argument), # Get the current day of month (1-31)
            'day_of_week': self.api.exchange.day_of_week(state_argument), # Get the current day of the week (1-7)
            'day_of_year': self.api.exchange.day_of_year(state_argument), # Get the current day of the year (1-366)
            'holiday_index': self.api.exchange.holiday_index(state_argument), # Gets a flag for the current day holiday type: 0 is no holiday, 1 is holiday type #1, etc.
            'hour': self.api.exchange.hour(state_argument), # Get the current hour of the simulation (0-23)
            'minutes': self.api.exchange.hour(state_argument), # Get the current minutes into the hour (1-60)
            'month': self.api.exchange.month(state_argument), # Get the current month of the simulation (1-12)
            'num_time_steps_in_hour': self.api.exchange.num_time_steps_in_hour(state_argument), # Returns the number of zone time steps in an hour, which is currently a constant value throughout a simulation.
            'year': self.api.exchange.year(state_argument), # Get the “current” year of the simulation, read from the EPW. All simulations operate at a real year, either user specified or automatically selected by EnergyPlus based on other data (start day of week + leap year option).
            'is_raining': self.api.exchange.is_raining(state_argument), # Gets a flag for whether the it is currently raining. The C API returns an integer where 1 is yes and 0 is no, this simply wraps that with a bool conversion.
            'sun_is_up': self.api.exchange.sun_is_up(state_argument), # Gets a flag for whether the sun is currently up. The C API returns an integer where 1 is yes and 0 is no, this simply wraps that with a bool conversion.
            # Gets the specified weather data at the specified hour and time step index within that hour
            'today_weather_albedo_at_time': self.api.exchange.today_weather_albedo_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_beam_solar_at_time': self.api.exchange.today_weather_beam_solar_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_diffuse_solar_at_time': self.api.exchange.today_weather_diffuse_solar_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_horizontal_ir_at_time': self.api.exchange.today_weather_horizontal_ir_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_is_raining_at_time': self.api.exchange.today_weather_is_raining_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_is_snowing_at_time': self.api.exchange.today_weather_is_snowing_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_liquid_precipitation_at_time': self.api.exchange.today_weather_liquid_precipitation_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_outdoor_barometric_pressure_at_time': self.api.exchange.today_weather_outdoor_barometric_pressure_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_outdoor_dew_point_at_time': self.api.exchange.today_weather_outdoor_dew_point_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_outdoor_dry_bulb_at_time': self.api.exchange.today_weather_outdoor_dry_bulb_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_outdoor_relative_humidity_at_time': self.api.exchange.today_weather_outdoor_relative_humidity_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_sky_temperature_at_time': self.api.exchange.today_weather_sky_temperature_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_wind_direction_at_time': self.api.exchange.today_weather_wind_direction_at_time(state_argument, hour, zone_time_step_number),
            'today_weather_wind_speed_at_time': self.api.exchange.today_weather_wind_speed_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_albedo_at_time': self.api.exchange.tomorrow_weather_albedo_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_beam_solar_at_time': self.api.exchange.tomorrow_weather_beam_solar_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_diffuse_solar_at_time': self.api.exchange.tomorrow_weather_diffuse_solar_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_horizontal_ir_at_time': self.api.exchange.tomorrow_weather_horizontal_ir_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_is_raining_at_time': self.api.exchange.tomorrow_weather_is_raining_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_is_snowing_at_time': self.api.exchange.tomorrow_weather_is_snowing_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_liquid_precipitation_at_time': self.api.exchange.tomorrow_weather_liquid_precipitation_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_outdoor_barometric_pressure_at_time': self.api.exchange.tomorrow_weather_outdoor_barometric_pressure_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_outdoor_dew_point_at_time': self.api.exchange.tomorrow_weather_outdoor_dew_point_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_outdoor_dry_bulb_at_time': self.api.exchange.tomorrow_weather_outdoor_dry_bulb_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_outdoor_relative_humidity_at_time': self.api.exchange.tomorrow_weather_outdoor_relative_humidity_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_sky_temperature_at_time': self.api.exchange.tomorrow_weather_sky_temperature_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_wind_direction_at_time': self.api.exchange.tomorrow_weather_wind_direction_at_time(state_argument, hour, zone_time_step_number),
            'tomorrow_weather_wind_speed_at_time': self.api.exchange.tomorrow_weather_wind_speed_at_time(state_argument, hour, zone_time_step_number),
        }
        variables = {}
        
        # Return the dictionary with variables names and output values of the methods used.
        include: List[str] = []
        parameters_keys = [key for key in parameter_methods.keys()]
        for paramater in parameters_keys:
            if self.observation_config[agent]["simulation_parameters"][paramater]:
                include.append(paramater)
        variables: Dict[str,Any] = {
            get_parameter_name(agent,paramater): parameter_methods[paramater] 
            for paramater 
            in include
        }
        return variables

    def get_zone_simulation_parameters_values(
        self,
        state_argument: c_void_p,
        agent: str,
        ) -> Dict[str,Any]:
        """Get the simulation parameters values defined in the simulation parameter list.

        Args:
            state_argument (): EnergyPlus state.
            simulation_parameter_list (List): List of the simulation parameters implemented in the observation space.

        Returns:
            Dict[str,int|float]: Dict of parameter names as keys and parameter values as values.
        """
        num_time_steps_in_hour = self.api.exchange.num_time_steps_in_hour(state_argument)
        # Dict with the variables names and methods as values.
        parameter_methods: Dict[str,Any] = {
            'system_time_step': self.api.exchange.system_time_step(state_argument), # Gets the current system time step value in EnergyPlus. The system time step is variable and fluctuates during the simulation.
            'zone_time_step': self.api.exchange.zone_time_step(state_argument), # Gets the current zone time step value in EnergyPlus. The zone time step is variable and fluctuates during the simulation.
            'zone_time_step_number': np.sin(2 * np.pi * (self.api.exchange.zone_time_step_number(state_argument)) / num_time_steps_in_hour), # The current zone time step index, from 1 to the number of zone time steps per hour
        }
        variables: Dict[str,Any] = {}
        
        # Return the dictionary with variables names and output values of the methods used.
        include: List[str] = []
        parameters_keys = [key for key in parameter_methods.keys()]
        for paramater in parameters_keys:
            if self.observation_config[agent]['zone_simulation_parameters'][paramater]:
                include.append(paramater)
        
        variables: Dict[str,Any] = {
            get_parameter_name(agent,paramater): parameter_methods[paramater] 
            for paramater 
            in include
        }
        return variables

    def get_weather_prediction(
        self,
        state_argument: c_void_p,
        agent: str
    ) -> Dict[str,Any]:
        assert isinstance(agent, str), "EnvironmentRunner: Agent must be a string."
        if not self.observation_config[agent]['use_one_day_weather_prediction']:
            return {}
        # Get timestep variables that are needed as input for some data_exchange methods.
        hour: int = self.api.exchange.hour(state_argument)
        
        prediction_variables_methods: Dict[str,Any] = {
            'today_weather_albedo_at_time': self.api.exchange.today_weather_albedo_at_time,
            'today_weather_beam_solar_at_time': self.api.exchange.today_weather_beam_solar_at_time,
            'today_weather_diffuse_solar_at_time': self.api.exchange.today_weather_diffuse_solar_at_time,
            'today_weather_horizontal_ir_at_time': self.api.exchange.today_weather_horizontal_ir_at_time,
            'today_weather_is_raining_at_time': self.api.exchange.today_weather_is_raining_at_time,
            'today_weather_is_snowing_at_time': self.api.exchange.today_weather_is_snowing_at_time,
            'today_weather_liquid_precipitation_at_time': self.api.exchange.today_weather_liquid_precipitation_at_time,
            'today_weather_outdoor_barometric_pressure_at_time': self.api.exchange.today_weather_outdoor_barometric_pressure_at_time,
            'today_weather_outdoor_dew_point_at_time': self.api.exchange.today_weather_outdoor_dew_point_at_time,
            'today_weather_outdoor_dry_bulb_at_time': self.api.exchange.today_weather_outdoor_dry_bulb_at_time,
            'today_weather_outdoor_relative_humidity_at_time': self.api.exchange.today_weather_outdoor_relative_humidity_at_time,
            'today_weather_sky_temperature_at_time': self.api.exchange.today_weather_sky_temperature_at_time,
            'today_weather_wind_direction_at_time': self.api.exchange.today_weather_wind_direction_at_time,
            'today_weather_wind_speed_at_time': self.api.exchange.today_weather_wind_speed_at_time,
            'tomorrow_weather_albedo_at_time': self.api.exchange.tomorrow_weather_albedo_at_time,
            'tomorrow_weather_beam_solar_at_time': self.api.exchange.tomorrow_weather_beam_solar_at_time,
            'tomorrow_weather_diffuse_solar_at_time': self.api.exchange.tomorrow_weather_diffuse_solar_at_time,
            'tomorrow_weather_horizontal_ir_at_time': self.api.exchange.tomorrow_weather_horizontal_ir_at_time,
            'tomorrow_weather_is_raining_at_time': self.api.exchange.tomorrow_weather_is_raining_at_time,
            'tomorrow_weather_is_snowing_at_time': self.api.exchange.tomorrow_weather_is_snowing_at_time,
            'tomorrow_weather_liquid_precipitation_at_time': self.api.exchange.tomorrow_weather_liquid_precipitation_at_time,
            'tomorrow_weather_outdoor_barometric_pressure_at_time': self.api.exchange.tomorrow_weather_outdoor_barometric_pressure_at_time,
            'tomorrow_weather_outdoor_dew_point_at_time': self.api.exchange.tomorrow_weather_outdoor_dew_point_at_time,
            'tomorrow_weather_outdoor_dry_bulb_at_time': self.api.exchange.tomorrow_weather_outdoor_dry_bulb_at_time,
            'tomorrow_weather_outdoor_relative_humidity_at_time': self.api.exchange.tomorrow_weather_outdoor_relative_humidity_at_time,
            'tomorrow_weather_sky_temperature_at_time': self.api.exchange.tomorrow_weather_sky_temperature_at_time,
            'tomorrow_weather_wind_direction_at_time': self.api.exchange.tomorrow_weather_wind_direction_at_time,
            'tomorrow_weather_wind_speed_at_time': self.api.exchange.tomorrow_weather_wind_speed_at_time,
        }
        
        variables: Dict[str,Any] = {}
        prediction_variables:Dict[str,bool] = self.observation_config[agent]['prediction_variables']
        for h in range(1, self.observation_config[agent]['weather_prediction_hours']+1, 1):
            # For each hour, the sigma value goes from a minimum error of zero to the value listed in sigma_max following a linear function:
            prediction_hour: int = hour + h
            if prediction_hour < 24:
                for key in prediction_variables.keys():
                    if prediction_variables[key]:
                        variables.update({
                            get_parameter_prediction_name(agent,key,h): prediction_variables_methods[f'today_weather_{key}_at_time'](state_argument,prediction_hour,1)
                        })
            else:
                prediction_hour_t = prediction_hour - 24
                for key in prediction_variables.keys():
                    if prediction_variables[key]:
                        variables.update({
                            get_parameter_prediction_name(agent,key,h): prediction_variables_methods[f'tomorrow_weather_{key}_at_time'](state_argument,prediction_hour_t,1)
                        })
        return variables
        
    def get_other_obs(
        self,
        agent: str
        ) -> Dict[str,Any]:
        """Get the static variable values defined in the static variable dict names and handles.

        Args:
            agent (str): The AgentID.

        Returns:
            Dict[str,Any]: Static value dict values for the actual timestep.
        """
        variables: Dict[str,Any] = {}
        variables.update({
            get_other_obs_name(agent,key): value 
            for key, value
            in self.observation_config[agent]["other_obs"].items()
        })
        self.observation_config[agent]["other_obs"]
        return variables
    
    def get_user_occupation_forecast(
        self,
        state_argument: c_void_p,
        agent: str
    ) -> Dict[str,Any]:
        """
        Get the user occupation forecast for the actual timestep.

        Args:
            state_argument (): EnergyPlus state.
            agent (str): The AgentID.

        Returns:
            Dict[str,Any]: User occupation forecast dict values for the actual timestep.
        """                
        # Get timestep variables that are needed as input for some data_exchange methods.
        if self.observation_config[agent]['user_occupation_function']:
            
            self.occupancy_next_timestep  = calculate_occupancy(
                self.api.exchange.hour(state_argument),
                self.api.exchange.day_of_month(state_argument),
                self.api.exchange.month(state_argument),
                self.api.exchange.year(state_argument),
                self.api.exchange.holiday_index(state_argument) > 0,
                self.observation_config[agent]['user_type'],
                self.observation_config[agent]['zone_type'],
                self.observation_config[agent]['confidence_level']
            )
        
        if self.observation_config[agent]['user_occupation_forecast']:
            variables: Dict[str,Any] = {}
            forecast_vector = calculate_occupancy_forecast(
                self.api.exchange.hour(state_argument),
                self.api.exchange.day_of_month(state_argument),
                self.api.exchange.month(state_argument),
                self.api.exchange.year(state_argument),
                self.observation_config[agent]['user_type'],
                self.observation_config[agent]['zone_type'],
                self.observation_config[agent]['occupation_prediction_hours'],
                self.observation_config[agent]['confidence_level'],
                self.observation_config[agent]['lambdaa']
            )
            
            for h in range(1, self.observation_config[agent]['occupation_prediction_hours']+1, 1):
                variables.update({get_user_occupation_forecast_name(agent,h): forecast_vector[h-1]})
                
            return variables
            
        else:
            return {}
        
        
    
    def _init_callback(self, state_argument: c_void_p) -> bool:
        """
        Initialize EnergyPlus handles and checks if simulation runtime is ready.
        
        Args:
            state_argument (): EnergyPlus state pointer. This is created with `api.state_manager.new_state()`.

        Returns:
            bool: True if the simulation is ready to perform actions.
        """
        if not self.init_handles:
            # print("Initializing handles...")
            self.init_handles = self._init_handles(state_argument)
        
        if self.api.exchange.warmup_flag(state_argument):
            self.initialized = False
            # print("Warmup phase still alive...")
            return self.initialized
        
        else:
            self.initialized = self.init_handles
            return self.initialized

    def _init_handles(self, state_argument: c_void_p) -> bool:
        """
        Initialize sensors/actuators handles to interact with during simulation.
        
        Args:
            state_argument (): EnergyPlus state pointer. This is created with `api.state_manager.new_state()`.

        Returns:
            bool: True if all handles were initialized successfully.
        """        
        if not self.init_handles:
            if not self.api.exchange.api_data_fully_ready(state_argument):
                return False
        
        self.internal_variables_and_handles["internal_actuators"][1].update({
            key: self.api.exchange.get_actuator_handle(state_argument, *actuator)
            for key, actuator in self.internal_variables_and_handles["internal_actuators"][0].items()
        })
        
        for agent in self.agents:
            self.agent_variables_and_handles[f"{agent}_variables"][1].update({
                key: self.api.exchange.get_variable_handle(state_argument, *actuator)
                for key, actuator in self.agent_variables_and_handles[f"{agent}_variables"][0].items()
            })
            self.agent_variables_and_handles[f"{agent}_internal_variables"][1].update({
                key: self.api.exchange.get_internal_variable_handle(state_argument, *var)
                for key, var in self.agent_variables_and_handles[f"{agent}_internal_variables"][0].items()
            })
            self.agent_variables_and_handles[f"{agent}_meters"][1].update({
                key: self.api.exchange.get_meter_handle(state_argument, meter)
                for key, meter in self.agent_variables_and_handles[f"{agent}_meters"][0].items()
            })
            self.agent_variables_and_handles[f"{agent}_actuators"][1].update({
                key: self.api.exchange.get_actuator_handle(state_argument, *actuator)
                for key, actuator in self.agent_variables_and_handles[f"{agent}_actuators"][0].items()
            })
            
            for handles in [
                self.agent_variables_and_handles[f"{agent}_variables"][1],
                self.agent_variables_and_handles[f"{agent}_internal_variables"][1],
                self.agent_variables_and_handles[f"{agent}_meters"][1],
                self.agent_variables_and_handles[f"{agent}_actuators"][1]
            ]:
                if any([v == -1 for v in handles.values()]):
                    available_data = self.api.exchange.list_available_api_data_csv(state_argument).decode('utf-8')
                    msg = (
                        f"EnvironmentRunner: \n" 
                        f"Some handles were not initialized correctly for agent {agent}.\n"
                        f"> handles with error: {handles}\n"
                        f"> available EnergyPlus API data: {available_data}"
                    )
                    logger.error(msg)
                    raise ValueError(msg)
        
        return True

    def stop(self) -> None:
        """
        Method to stop EnergyPlus simulation and joint the threads.
        """
        # print("Stopping EnergyPlus simulation...")
        if not self.simulation_complete:
            self.simulation_complete = True
        
        try:
            if self.energyplus_state is not None:
                self.api.runtime.stop_simulation(self.energyplus_state)
        except Exception as e:
            logger.warning(f"EnvironmentRunner: Error stopping EnergyPlus simulation: {e}")
        
        self._flush_queues()
        
        if self.energyplus_exec_thread is not None:
            try:
                self.energyplus_exec_thread.join(timeout=self.energyplus_exec_thread_timeout)
            except Exception as e:
                logger.warning(f"EnvironmentRunner: Error joining EnergyPlus thread: {e}")
        
        self.energyplus_exec_thread = None
        self.first_observation = True
        
        try:
            self.api.runtime.clear_callbacks()
        except Exception as e:
            logger.warning(f"EnvironmentRunner: Error clearing callbacks: {e}")
        
        if self.energyplus_state is not None:
            try:
                self.api.state_manager.delete_state(self.energyplus_state)
            except Exception as e:
                logger.warning(f"EnvironmentRunner: Error deleting EnergyPlus state: {e}")
            finally:
                self.energyplus_state = None

    def failed(self) -> bool:
        """
        This method tells if a EnergyPlus simulations was finished successfully or not.

        Returns:
            bool: Boolean value of the success of the simulation.
        """
        return self.sim_results != 0

    def make_eplus_args(self) -> List[str|bytes]:
        """
        Make command line arguments to pass to EnergyPlus.
        
        Return:
            List[str]: List of arguments to pass to EnergyPlusEnv.
        """
        eplus_args: List[str|bytes] = ["-r"] if self.generals_config.get("csv", False) else []
        eplus_args += [
            "-w",
            self.generals_config["epw_path"],
            "-d",
            f"{self.generals_config['output_path']}/episode-{self.episode:08}-{os.getpid():05}-{self.unique_id}" if not self.generals_config['evaluation'] else f"{self.generals_config['output_path']}/evaluation-episode-{self.episode:08}-{os.getpid():05}-{self.unique_id}",
            self.generals_config["epjson_path"]
        ]
        return eplus_args
    
    def _flush_queues(self):
        """
        Method to liberate the space in the different queue objects.
        """
        for q in [self.obs_queue, self.act_queue, self.infos_queue]:
            while not q.empty():
                q.get()



"""
Specification for the action space and actuators
===========================================================
This module defines the ``ActionSpec`` class, which is used to specify the 
configuration of action space and actuators for agents in reinforcement 
learning environments.
"""
from typing import Dict, Tuple, Any, Optional

from eprllib import logger

class ActionSpec:
    """
    ActionSpec is the base class for an action specification to safe configuration of the object.
    """
    actuators: Optional[Dict[str, Tuple[str, str, str]]] = None
    
    def __init__(
        self,
        actuators: Optional[Dict[str, Tuple[str, str, str]]] = None,
    ):
        """
        Construction method.
        
        Args:
        
            actuators (Dict[str, Tuple[str, str, str]]): Actuators are the way that users modify the program at 
                runtime using custom logic and calculations. Not every variable inside EnergyPlus can be 
                actuated. This is intentional, because opening that door could allow the program to run at 
                unrealistic conditions, with flow imbalances or energy imbalances, and many other possible problems.
                Instead, a specific set of items are available to actuate, primarily control functions, 
                flow requests, and environmental boundary conditions. These actuators, when used in conjunction 
                with the runtime API and data exchange variables, allow a user to read data, make decisions and 
                perform calculations, then actuate control strategies for subsequent time steps.
                Actuator functions are similar, but not exactly the same, as for variables. An actuator
                handle/ID is still looked up, but it takes the actuator type, component name, and control
                type, since components may have more than one control type available for actuation. The
                actuator can then be “actuated” by calling a set-value function, which overrides an internal
                value, and informs EnergyPlus that this value is currently being externally controlled. To
                allow EnergyPlus to resume controlling that value, there is an actuator reset function as well.
                One agent can manage several actuators.
        
        Raises:
        
            ValueError: If the actuators are not defined as a list of tuples of 3 elements.
            KeyError: If an invalid key is provided when setting an item.
            
        """
        self.actuators = actuators
        
    
    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)
    

    def __setitem__(self, key: str, value: Any) -> None:
        valid_keys = self.__dict__.keys()
        if key not in valid_keys:
            msg = f"ActionSpec: Invalid key: {key}."
            logger.error(msg)
            raise KeyError(msg)
        setattr(self, key, value)
        
        
    def build(self) -> Dict[str, Any]:
        """
        This method is used to build the ActionSpec object.
        """
        if self.actuators is None:
            logger.info("ActionSpec: No actuators provided.")
            self.actuators = {}
        
        if not isinstance(self.actuators, dict): # type: ignore
            msg = f"ActionSpec: The actuators must be defined as a dict with actuator names as keys and tuples of 3 elements but {type(self.actuators)} was given."
            logger.error(msg)
            raise ValueError(msg)
        
        # Check that the actuators are defined as a list of tuples of 3 elements.
        for actuator in self.actuators.values():
            if len(actuator) != 3:
                msg = f"ActionSpec: The actuators must be defined as a dict with actuator names as keys and tuples of 3 elements but {len(actuator)} was given."
                logger.error(msg)
                raise ValueError(msg)
        
        return vars(self)
    